# init the package folder
__all__ = [
    'auxilary_frames', 'draw_frames', 'dynamic_timings', 'frame_position', 'get_time_offset',
    'mmrhist', 'randoms', 'split_frames']

from .mmrhist import (
    auxilary_frames,
    draw_frames,
    dynamic_timings,
    frame_position,
    get_time_offset,
    mmrhist,
    randoms,
    split_frames,
)

# from .pviews import video_frm, video_dyn



from os import getenv
from pathlib import Path

from pytest import fixture, skip

HOME = Path(getenv("DATA_ROOT", "~")).expanduser()


@fixture(scope="session")
def folder_in():
    Ab_PET_mMR_test = HOME / "Ab_PET_mMR_test"
    if not Ab_PET_mMR_test.is_dir():
        skip(f"""Cannot find Ab_PET_mMR_test in ${{DATA_ROOT:-~}} ({HOME}).
Try running `python -m tests` to download it.
""")
    return Ab_PET_mMR_test


@fixture(scope="session")
def folder_ref(folder_in):
    Ab_PET_mMR_ref = folder_in / "testing_reference" / "Ab_PET_mMR_ref"
    if not Ab_PET_mMR_ref.is_dir():
        skip(f"""Cannot find Ab_PET_mMR_ref in
${{DATA_ROOT:-~}}/testing_reference ({HOME}/testing_reference).
Try running `python -m tests` to download it.
""")
    return Ab_PET_mMR_ref



import errno
import logging
from collections.abc import Iterable
from os import fspath

import numpy as np
import pytest
from filelock import FileLock

from niftypet import nimpa, nipet

# segmentation/parcellation for PVC, with unique regions numbered from 0 onwards
pvcroi = []
pvcroi.append([66, 67] + list(range(81, 95)))             # white matter
pvcroi.append([36])                                       # brain stem
pvcroi.append([35])                                       # pons
pvcroi.append([39, 40, 72, 73, 74])                       # cerebellum GM
pvcroi.append([41, 42])                                   # cerebellum WM
pvcroi.append([48, 49])                                   # hippocampus
pvcroi.append([167, 168])                                 # posterior cingulate gyrus
pvcroi.append([139, 140])                                 # middle cingulate gyrus
pvcroi.append([101, 102])                                 # anterior cingulate gyrus
pvcroi.append([169, 170])                                 # precuneus
pvcroi.append([32, 33])                                   # amygdala
pvcroi.append([37, 38])                                   # caudate
pvcroi.append([56, 57])                                   # pallidum
pvcroi.append([58, 59])                                   # putamen
pvcroi.append([60, 61])                                   # thalamus
pvcroi.append([175, 176, 199, 200])                       # parietal without precuneus
pvcroi.append([133, 134, 155, 156, 201, 202, 203, 204])   # temporal
pvcroi.append([4, 5, 12, 16, 43, 44, 47, 50, 51, 52, 53]) # CSF
pvcroi.append([24, 31, 62, 63, 70, 76, 77, 96, 97])       # basal ganglia + optic chiasm

# remaining neocortex
pvcroi.append(
    list(range(103, 110 + 1)) + list(range(113, 126 + 1)) + list(range(129, 130 + 1)) +
    list(range(135, 138 + 1)) + list(range(141, 154 + 1)) + list(range(157, 158 + 1)) +
    list(range(161, 166 + 1)) + list(range(171, 174 + 1)) + list(range(177, 188 + 1)) +
    list(range(191, 198 + 1)) + list(range(205, 208 + 1)))

# expected %error for static (SUVr) and PVC reconstructions
emape_basic = 0.1
emape_algnd = {"pet": 3.0, "pos": 0.1, "trm": 3.0, "pvc": 3.0, "hmu": 0.01, "omu": 3.0}


@pytest.fixture(scope="session")
def mMRpars():
    params = nipet.get_mmrparams()
    # params["Cnt"]["VERBOSE"] = True
    params["Cnt"]["LOG"] = logging.INFO
    return params


@pytest.fixture(scope="session")
def datain(mMRpars, folder_in):
    return nipet.classify_input(folder_in, mMRpars)


@pytest.fixture(scope="session")
def muhdct(mMRpars, datain, tmp_path_factory, worker_id):
    tmp_path = tmp_path_factory.getbasetemp()

    if worker_id == "master": # not xdist, auto-reuse
        opth = str(tmp_path / "muhdct")
        return nipet.hdw_mumap(datain, [1, 2, 4], mMRpars, outpath=opth, use_stored=True)

    opth = str(tmp_path.parent / "muhdct")
    flock = FileLock(opth + ".lock")
    with flock.acquire(poll_intervall=0.5): # xdist, force auto-reuse via flock
        return nipet.hdw_mumap(datain, [1, 2, 4], mMRpars, outpath=opth, use_stored=True)


@pytest.fixture(scope="session")
def refimg(folder_ref):
    # predetermined structure of the reference folder
    basic = folder_ref / "basic"
    spm = folder_ref / "dyn_aligned" / "spm"
    niftyreg = folder_ref / "dyn_aligned" / "niftyreg"
    refpaths = {
        "histo": {"p": 1570707830, "d": 817785422}, "basic": {
            "pet": basic / "17598013_t-3000-3600sec_itr-4_suvr.nii.gz",
            "omu": basic / "mumap-from-DICOM_no-alignment.nii.gz",
            "hmu": basic / "hardware_umap.nii.gz"},
        "aligned": {
            "spm": {
                "hmu": spm / "hardware_umap.nii.gz",
                "omu": spm / "mumap-PCT-aligned-to_t0-3600_AC.nii.gz",
                "pos": spm / "17598013_t0-3600sec_itr2_AC-UTE.nii.gz",
                "pet": spm / "17598013_nfrm-2_itr-4.nii.gz",
                "trm": spm / "17598013_nfrm-2_itr-4_trimmed-upsampled-scale-2.nii.gz",
                "pvc": spm / "17598013_nfrm-2_itr-4_trimmed-upsampled-scale-2_PVC.nii.gz"},
            "niftyreg": {
                "hmu": niftyreg / "hardware_umap.nii.gz",
                "omu": niftyreg / "mumap-PCT-aligned-to_t0-3600_AC.nii.gz",
                "pos": niftyreg / "17598013_t0-3600sec_itr2_AC-UTE.nii.gz",
                "pet": niftyreg / "17598013_nfrm-2_itr-4.nii.gz",
                "trm": niftyreg / "17598013_nfrm-2_itr-4_trimmed-upsampled-scale-2.nii.gz",
                "pvc": niftyreg / "17598013_nfrm-2_itr-4_trimmed-upsampled-scale-2_PVC.nii.gz"}}}

    testext = {
        "basic": {
            "pet": "static reconstruction with unaligned UTE mu-map",
            "hmu": "hardware mu-map for the static unaligned reconstruction",
            "omu": "object mu-map for the static unaligned reconstruction"}, "aligned": {
                "hmu": "hardware mu-map for the 2-frame aligned reconstruction",
                "omu": "object mu-map for the 2-frame aligned reconstruction",
                "pos": "AC reconstruction for positioning (full acquisition used)",
                "pet": "2-frame scan with aligned UTE mu-map",
                "trm": "trimming post reconstruction", "pvc": "PVC post reconstruction"}}

    # check basic files
    for k, v in refpaths["basic"].items():
        if not v.is_file():
            raise FileNotFoundError(errno.ENOENT, f"{k}: {v}")

    # check reg tools: niftyreg and spm
    for r, frefs in refpaths["aligned"].items():
        for k, v in frefs.items():
            if not v.is_file():
                raise FileNotFoundError(errno.ENOENT, f"{k}[{r}]: {v}")

    return refpaths, testext


def test_histogramming(mMRpars, datain, refimg, tmp_path):
    refpaths, _ = refimg
    opth = str(tmp_path / "histogramming")
    hst = nipet.mmrhist(datain, mMRpars, outpath=opth, store=True)

    # prompt counts: head curve & sinogram
    assert np.sum(hst["phc"]) == np.sum(hst["psino"])
    # delayed counts: head curve & sinogram
    assert np.sum(hst["dhc"]) == np.sum(hst["dsino"])
    # prompt counts: head curve & reference
    assert np.sum(hst["phc"]) == refpaths["histo"]["p"]
    # delayed counts: head curve & reference
    assert np.sum(hst["dhc"]) == refpaths["histo"]["d"]


def test_basic_reconstruction(mMRpars, datain, muhdct, refimg, tmp_path):
    refpaths, testext = refimg
    opth = str(tmp_path / "basic_reconstruction")

    muodct = nipet.obj_mumap(datain, mMRpars, outpath=opth, store=True)
    recon = nipet.mmrchain(
        datain,
        mMRpars,
        frames=["timings", [3000, 3600]],
        mu_h=muhdct,
        mu_o=muodct,
        itr=4,
        fwhm=0.0,
        outpath=opth,
        fcomment="_suvr",
        store_img=True,
        store_img_intrmd=True,
    )

    testout = {"pet": recon["fpet"], "hmu": muhdct["im"], "omu": muodct["im"]}
    for k in testext["basic"]:
        diff = nimpa.imdiff(fspath(refpaths["basic"][k]), testout[k], verbose=True, plot=False)
        assert diff["mape"] <= emape_basic, testext["basic"][k]


@pytest.mark.parametrize("reg_tool", ["niftyreg", "spm"])
def test_aligned_reconstruction(reg_tool, mMRpars, datain, muhdct, refimg, tmp_path):
    refpaths, testext = refimg
    opth = str(tmp_path / "aligned_reconstruction")

    muopct = nipet.align_mumap(
        datain,
        mMRpars,
        outpath=opth,
        t0=0,
        t1=3600,
        reg_tool=reg_tool,
        store=True,
        itr=2,
        petopt="ac",
        fcomment="",
        musrc="pct",
    )
    # matshow(muopct['im'][60,:,:] + muhdct['im'][60,:,:])
    recon = nipet.mmrchain(
        datain,
        mMRpars,
        frames=["timings", [80, 260], [3000, 3600]],
        mu_h=muhdct,
        mu_o=muopct,
        itr=4,
        fwhm=0.0,
        outpath=opth,
        trim=True,
        pvcroi=pvcroi,
        pvcreg_tool=reg_tool,
        store_img=True,
        store_img_intrmd=True,
    )

    testout = {
        "pet": recon["fpet"], "hmu": muhdct["im"], "omu": muopct["im"], "pos": muopct["fpet"],
        "trm": recon["trimmed"]["fpet"], "pvc": recon["trimmed"]["fpvc"]}
    for k in testext["aligned"]:
        diff = nimpa.imdiff(fspath(refpaths["aligned"][reg_tool][k]), testout[k], verbose=True,
                            plot=False)
        err = diff["mape"] <= emape_algnd[k]
        assert (all(err) if isinstance(err, Iterable) else err), testext["aligned"][k]



"""Image functions for PET data reconstruction and processing."""

import glob
import logging
import math
import multiprocessing
import os
import re
import shutil
from subprocess import run

import nibabel as nib
import numpy as np
import pydicom as dcm

from niftypet import nimpa

from .. import mmraux
from .. import resources as rs

log = logging.getLogger(__name__)
OFFSET_DEFAULT = np.array([0., 0., 0.])
ct_nans = -1024

# ==================================================================================
# IMAGE ROUTINES
# ==================================================================================


def convert2e7(img, Cnt):
    '''Convert GPU optimised image to Siemens/E7 image shape (127,344,344).'''

    margin = (Cnt['SO_IMX'] - Cnt['SZ_IMX']) // 2

    # permute the dims first
    imo = np.transpose(img, (2, 0, 1))

    nvz = img.shape[2]

    # > get the x-axis filler and apply it
    filler = np.zeros((nvz, Cnt['SZ_IMY'], margin), dtype=np.float32)
    imo = np.concatenate((filler, imo, filler), axis=2)

    # > get the y-axis filler and apply it
    filler = np.zeros((nvz, margin, Cnt['SO_IMX']), dtype=np.float32)
    imo = np.concatenate((filler, imo, filler), axis=1)
    return imo


def convert2dev(im, Cnt):
    '''Reshape Siemens/E7 (default) image for optimal GPU execution.'''
    if im.shape[1] != Cnt['SO_IMY'] or im.shape[2] != Cnt['SO_IMX']:
        raise ValueError('e> input image array is not of the correct Siemens shape.')

    if 'rSZ_IMZ' in Cnt and im.shape[0] != Cnt['rSZ_IMZ']:
        log.warning('the axial number of voxels does not match the reduced rings.')
    elif 'rSZ_IMZ' not in Cnt and im.shape[0] != Cnt['SZ_IMZ']:
        log.warning('the axial number of voxels does not match the rings.')

    im_sqzd = np.zeros((im.shape[0], Cnt['SZ_IMY'], Cnt['SZ_IMX']), dtype=np.float32)
    margin = int((Cnt['SO_IMX'] - Cnt['SZ_IMX']) / 2)
    margin_ = -margin
    if margin == 0:
        margin = None
        margin_ = None

    im_sqzd = im[:, margin:margin_, margin:margin_]
    im_sqzd = np.transpose(im_sqzd, (1, 2, 0))
    return im_sqzd


def cropxy(im, imsize, datain, Cnt, store_pth=''):
    '''
    Crop image transaxially to the size in tuple <imsize>.
    Return the image and the affine matrix.
    '''
    if not imsize[0] % 2 == 0 and not imsize[1] % 2 == 0:
        log.error('image size has to be an even number!')
        return None

    # cropping indexes
    i0 = int((Cnt['SO_IMX'] - imsize[0]) / 2)
    i1 = int((Cnt['SO_IMY'] + imsize[1]) / 2)

    B = image_affine(datain, Cnt, gantry_offset=False)
    B[0, 3] -= 10 * Cnt['SO_VXX'] * i0
    B[1, 3] += 10 * Cnt['SO_VXY'] * (Cnt['SO_IMY'] - i1)

    cim = im[:, i0:i1, i0:i1]

    if store_pth != '':
        nimpa.array2nii(cim[::-1, ::-1, :], B, store_pth, descrip='cropped')
        log.info('saved cropped image to:\n{}'.format(store_pth))

    return cim, B


def image_affine(datain, Cnt, gantry_offset=False):
    '''Creates a blank reference image, to which another image will be resampled'''

    # ------get necessary data for -----
    # gantry offset
    if gantry_offset:
        goff, tpo = mmraux.lm_pos(datain, Cnt)
    else:
        goff = np.zeros((3))
    vbed, hbed = mmraux.vh_bedpos(datain, Cnt)

    if 'rNRNG' in Cnt and 'rSO_IMZ' in Cnt:
        imz = Cnt['rSO_IMZ']
    else:
        imz = Cnt['SO_IMZ']

    # create a reference empty mu-map image
    B = np.diag(np.array([-10 * Cnt['SO_VXX'], 10 * Cnt['SO_VXY'], 10 * Cnt['SO_VXZ'], 1]))
    B[0, 3] = 10 * (.5 * Cnt['SO_IMX'] * Cnt['SO_VXX'] + goff[0])
    B[1, 3] = 10 * ((-.5 * Cnt['SO_IMY'] + 1) * Cnt['SO_VXY'] - goff[1])
    B[2, 3] = 10 * ((-.5 * imz + 1) * Cnt['SO_VXZ'] - goff[2] + hbed)
    # -------------------------------------------------------------------------------------
    return B


def getmu_off(mu, Cnt, Offst=OFFSET_DEFAULT):

    # phange the shape to 3D
    mu.shape = (Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX'])

    # -------------------------------------------------------------------------
    # CORRECT THE MU-MAP for GANTRY OFFSET
    # -------------------------------------------------------------------------
    Cim = {
        'VXSOx': 0.208626, 'VXSOy': 0.208626, 'VXSOz': 0.203125, 'VXNOx': 344, 'VXNOy': 344,
        'VXNOz': 127, 'VXSRx': 0.208626, 'VXSRy': 0.208626, 'VXSRz': 0.203125, 'VXNRx': 344,
        'VXNRy': 344, 'VXNRz': 127}
    # priginal image offset
    Cim['OFFOx'] = -0.5 * Cim['VXNOx'] * Cim['VXSOx']
    Cim['OFFOy'] = -0.5 * Cim['VXNOy'] * Cim['VXSOy']
    Cim['OFFOz'] = -0.5 * Cim['VXNOz'] * Cim['VXSOz']
    # pesampled image offset
    Cim['OFFRx'] = -0.5 * Cim['VXNRx'] * Cim['VXSRx']
    Cim['OFFRy'] = -0.5 * Cim['VXNRy'] * Cim['VXSRy']
    Cim['OFFRz'] = -0.5 * Cim['VXNRz'] * Cim['VXSRz']
    # pransformation matrix
    A = np.array(
        [[1., 0., 0., Offst[0]], [0., 1., 0., Offst[1]], [0., 0., 1., Offst[2]], [0., 0., 0., 1.]],
        dtype=np.float32)
    # ppply the gantry offset to the mu-map
    mur = nimpa.prc.improc.resample(mu, A, Cim)
    return mur


def getinterfile_off(fmu, Cnt, Offst=OFFSET_DEFAULT):
    '''
    Return the floating point mu-map in an array from Interfile,
    accounting for image offset (does slow interpolation).
    '''
    # pead the image file
    f = open(fmu, 'rb')
    mu = np.fromfile(f, np.float32)
    f.close()

    # save_im(mur, Cnt, os.path.dirname(fmu) + '/mur.nii')
    # -------------------------------------------------------------------------
    mur = getmu_off(mu, Cnt)
    # > create GPU version of the mu-map
    murs = convert2dev(mur, Cnt)
    # > number of voxels
    nvx = mu.shape[0]
    # > get the basic stats
    mumax = np.max(mur)
    mumin = np.min(mur)
    # > number of voxels greater than 10% of max image value
    n10mx = np.sum(mur > 0.1 * mumax)
    # > return image dictionary with the image itself and some other stats
    mu_dct = {'im': mur, 'ims': murs, 'max': mumax, 'min': mumin, 'nvx': nvx, 'n10mx': n10mx}
    return mu_dct


def getinterfile(fim, Cnt):
    '''Return the floating point image file in an array from an Interfile file.'''
    # pead the image file
    f = open(fim, 'rb')
    im = np.fromfile(f, np.float32)
    f.close()

    # pumber of voxels
    nvx = im.shape[0]
    # phange the shape to 3D
    im.shape = (Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX'])

    # pet the basic stats
    immax = np.max(im)
    immin = np.min(im)

    # pumber of voxels greater than 10% of max image value
    n10mx = np.sum(im > 0.1 * immax)

    # peorganise the image for optimal gpu execution
    im_sqzd = convert2dev(im, Cnt)

    # peturn image dictionary with the image itself and some other stats
    im_dct = {'im': im, 'ims': im_sqzd, 'max': immax, 'min': immin, 'nvx': nvx, 'n10mx': n10mx}

    return im_dct


# define uniform cylinder


def get_cylinder(Cnt, rad=25, xo=0, yo=0, unival=1, gpu_dim=False):
    """
    Outputs image with a uniform cylinder of
    intensity = unival, radius = rad, and transaxial centre (xo, yo)
    """
    imdsk = np.zeros((1, Cnt['SO_IMX'], Cnt['SO_IMY']), dtype=np.float32)
    for t in np.arange(0, math.pi, math.pi / (2*360)):
        x = xo + rad * math.cos(t)
        y = yo + rad * math.sin(t)
        yf = np.arange(-y + 2*yo, y, Cnt['SO_VXY'] / 2)
        v = np.int32(.5 * Cnt['SO_IMX'] - np.ceil(yf / Cnt['SO_VXY']))
        u = np.int32(.5 * Cnt['SO_IMY'] + np.floor(x / Cnt['SO_VXY']))
        imdsk[0, v, u] = unival
    if 'rSO_IMZ' in Cnt:
        nvz = Cnt['rSO_IMZ']
    else:
        nvz = Cnt['SO_IMZ']
    imdsk = np.repeat(imdsk, nvz, axis=0)
    if gpu_dim: imdsk = convert2dev(imdsk, Cnt)
    return imdsk


def hu2mu(im):
    '''HU units to 511keV PET mu-values'''
    # convert nans to -1024 for the HU values only
    im[np.isnan(im)] = ct_nans
    # constants
    muwater = 0.096
    mubone = 0.172
    rhowater = 0.158
    rhobone = 0.326
    uim = np.zeros(im.shape, dtype=np.float32)
    uim[im <= 0] = muwater * (1 + im[im <= 0] * 1e-3)
    uim[im > 0] = muwater * (1 + im[im > 0] * 1e-3 * rhowater / muwater * (mubone-muwater) /
                             (rhobone-rhowater))
    # remove negative values
    uim[uim < 0] = 0
    return uim


# =====================================================================================
# object/patient mu-map resampling to NIfTI
# better use dcm2niix
def mudcm2nii(datain, Cnt):
    '''DICOM mu-map to NIfTI'''
    mu, pos, ornt = nimpa.dcm2im(datain['mumapDCM'])
    mu *= 0.0001
    A = pos['AFFINE']
    A[0, 0] *= -1
    A[0, 3] *= -1
    A[1, 3] += A[1, 1]
    nimpa.array2nii(mu[:, ::-1, :], A,
                    os.path.join(os.path.dirname(datain['mumapDCM']), 'mu.nii.gz'))

    # ------get necessary data for creating a blank reference image (to which resample)-----
    # gantry offset
    goff, tpo = mmraux.lm_pos(datain, Cnt)
    ihdr, csainfo = mmraux.hdr_lm(datain)
    # ptart horizontal bed position
    p = re.compile(r'start horizontal bed position.*\d{1,3}\.*\d*')
    m = p.search(ihdr)
    fi = ihdr[m.start():m.end()].find('=')
    hbedpos = 0.1 * float(ihdr[m.start() + fi + 1:m.end()])

    B = np.diag(np.array([-10 * Cnt['SO_VXX'], 10 * Cnt['SO_VXY'], 10 * Cnt['SO_VXZ'], 1]))
    B[0, 3] = 10 * (.5 * Cnt['SO_IMX'] * Cnt['SO_VXX'] + goff[0])
    B[1, 3] = 10 * ((-.5 * Cnt['SO_IMY'] + 1) * Cnt['SO_VXY'] - goff[1])
    B[2, 3] = 10 * ((-.5 * Cnt['SO_IMZ'] + 1) * Cnt['SO_VXZ'] - goff[2] + hbedpos)
    im = np.zeros((Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']), dtype=np.float32)
    nimpa.array2nii(im, B, os.path.join(os.path.dirname(datain['mumapDCM']), 'muref.nii.gz'))
    # -------------------------------------------------------------------------------------
    fmu = os.path.join(os.path.dirname(datain['mumapDCM']), 'mu_r.nii.gz')
    if os.path.isfile(Cnt['RESPATH']):
        run([
            Cnt['RESPATH'], '-ref',
            os.path.join(os.path.dirname(datain['mumapDCM']), 'muref.nii.gz'), '-flo',
            os.path.join(os.path.dirname(datain['mumapDCM']), 'mu.nii.gz'), '-res', fmu, '-pad',
            '0'])
    else:
        log.error('path to resampling executable is incorrect!')
        raise IOError('Error launching NiftyReg for image resampling.')

    return fmu


def obj_mumap(
    datain,
    params=None,
    outpath='',
    comment='',
    store=False,
    store_npy=False,
    gantry_offset=True,
    del_auxilary=True,
):
    '''Get the object mu-map from DICOM images'''
    if params is None:
        params = {}

    # three ways of passing scanner constants <Cnt> are here decoded
    if 'Cnt' in params:
        Cnt = params['Cnt']
    elif 'SO_IMZ' in params:
        Cnt = params
    else:
        Cnt = rs.get_mmr_constants()

    # output folder
    if outpath == '':
        fmudir = os.path.join(datain['corepath'], 'mumap-obj')
    else:
        fmudir = os.path.join(outpath, 'mumap-obj')
    nimpa.create_dir(fmudir)

    # > ref file name
    fmuref = os.path.join(fmudir, 'muref.nii.gz')

    # > ref affine
    B = image_affine(datain, Cnt, gantry_offset=gantry_offset)

    # > ref image (blank)
    im = np.zeros((Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']), dtype=np.float32)

    # > store ref image
    nimpa.array2nii(im, B, fmuref)

    # check if the object dicom files for MR-based mu-map exists
    if 'mumapDCM' not in datain or not os.path.isdir(datain['mumapDCM']):
        log.error('DICOM folder for the mu-map does not exist.')
        return None

    fnii = 'converted-from-object-DICOM_'
    tstmp = nimpa.time_stamp(simple_ascii=True)

    # find residual(s) from previous runs and delete them
    resdcm = glob.glob(os.path.join(fmudir, '*' + fnii + '*.nii*'))
    for d in resdcm:
        os.remove(d)

    # convert the DICOM mu-map images to nii
    run([Cnt['DCM2NIIX'], '-f', fnii + tstmp, '-o', fmudir, datain['mumapDCM']])
    # piles for the T1w, pick one:
    fmunii = glob.glob(os.path.join(fmudir, '*' + fnii + tstmp + '*.nii*'))[0]
    # fmunii = glob.glob( os.path.join(datain['mumapDCM'], '*converted*.nii*') )
    # fmunii = fmunii[0]

    # the converted nii image resample to the reference size
    fmu = os.path.join(fmudir, comment + 'mumap_tmp.nii.gz')
    if os.path.isfile(Cnt['RESPATH']):
        cmd = [Cnt['RESPATH'], '-ref', fmuref, '-flo', fmunii, '-res', fmu, '-pad', '0']
        if log.getEffectiveLevel() > logging.INFO:
            cmd.append('-voff')
        run(cmd)
    else:
        log.error('path to resampling executable is incorrect!')
        raise IOError('Path to executable is incorrect!')

    nim = nib.load(fmu)
    # get the affine transform
    A = nim.get_sform()
    mu = nim.get_fdata(dtype=np.float32)
    mu = np.transpose(mu[:, ::-1, ::-1], (2, 1, 0))
    # convert to mu-values
    mu = np.float32(mu) / 1e4
    mu[mu < 0] = 0

    # > return image dictionary with the image itself and some other stats
    mu_dct = {'im': mu, 'affine': A}
    if not del_auxilary:
        mu_dct['fmuref'] = fmuref

    # > store the mu-map if requested
    if store_npy:
        # to numpy array
        fnp = os.path.join(fmudir, "mumap-from-DICOM.npz")
        np.savez(fnp, mu=mu, A=A)

    if store:
        # with this file name
        fmumap = os.path.join(fmudir, 'mumap-from-DICOM_no-alignment' + comment + '.nii.gz')
        nimpa.array2nii(mu[::-1, ::-1, :], A, fmumap)
        mu_dct['fim'] = fmumap

    if del_auxilary:
        os.remove(fmuref)
        os.remove(fmunii)
        os.remove(fmu)

        if [f for f in os.listdir(fmudir)
                if not f.startswith('.') and not f.endswith('.json')] == []:
            shutil.rmtree(fmudir)

    return mu_dct


# ================================================================================
# pCT/UTE MU-MAP ALIGNED
# --------------------------------------------------------------------------------


def align_mumap(
    datain,
    scanner_params=None,
    outpath='',
    reg_tool='niftyreg',
    use_stored=False,
    hst=None,
    t0=0,
    t1=0,
    itr=2,
    faff='',
    fpet='',
    fcomment='',
    store=False,
    store_npy=False,
    petopt='ac',
    musrc='ute',         # another option is pct for mu-map source
    ute_name='UTE2',
    del_auxilary=True,
    verbose=False,
):
    '''
    Align the a pCT or MR-derived mu-map to a PET image reconstructed to chosen
    specifications (e.g., with/without attenuation and scatter corrections)

    use_sotred only works if hst or t0/t1 given but not when faff.
    '''
    if scanner_params is None:
        scanner_params = {}

    # > output folder
    if outpath == '':
        opth = os.path.join(datain['corepath'], 'mumap-obj')
    else:
        opth = os.path.join(outpath, 'mumap-obj')

    # > create the folder, if not existent
    nimpa.create_dir(opth)

    # > get the timing of PET if affine not given
    if faff == '' and hst is not None and isinstance(hst, dict) and 't0' in hst:
        t0 = hst['t0']
        t1 = hst['t1']

    # > file name for the output mu-map
    fnm = 'mumap-' + musrc.upper()

    # > output dictionary
    mu_dct = {}

    # ---------------------------------------------------------------------------
    # > used stored if requested
    if use_stored:
        fmu_stored = fnm + '-aligned-to_t'\
                     + str(t0)+'-'+str(t1)+'_'+petopt.upper()\
                     + fcomment
        fmupath = os.path.join(opth, fmu_stored + '.nii.gz')

        if os.path.isfile(fmupath):
            mudct_stored = nimpa.getnii(fmupath, output='all')
            # > create output dictionary
            mu_dct['im'] = mudct_stored['im']
            mu_dct['affine'] = mudct_stored['affine']
            # pu_dct['faff'] = faff
            return mu_dct
    # ---------------------------------------------------------------------------

    # > tmp folder for not aligned mu-maps
    tmpdir = os.path.join(opth, 'tmp')
    nimpa.create_dir(tmpdir)

    # > three ways of passing scanner constants <Cnt> are here decoded
    if 'Cnt' in scanner_params:
        Cnt = scanner_params['Cnt']
    elif 'SO_IMZ' in scanner_params:
        Cnt = scanner_params
    else:
        Cnt = rs.get_mmr_constants()

    # > if affine not provided histogram the LM data for recon and registration
    if not os.path.isfile(faff):
        from niftypet.nipet.prj import mmrrec

        # -histogram the list data if needed
        if hst is None:
            from niftypet.nipet import mmrhist
            if 'txLUT' in scanner_params:
                hst = mmrhist(datain, scanner_params, t0=t0, t1=t1)
            else:
                raise ValueError('Full scanner are parameters not provided\
                     but are required for histogramming.')

    # ========================================================
    # -get hardware mu-map
    if 'hmumap' in datain and os.path.isfile(datain['hmumap']):
        muh = np.load(datain['hmumap'], allow_pickle=True)["hmu"]
        (log.info if verbose else log.debug)('loaded hardware mu-map from file:\n{}'.format(
            datain['hmumap']))
    elif outpath != '':
        hmupath = os.path.join(outpath, "mumap-hdw", "hmumap.npz")
        if os.path.isfile(hmupath):
            muh = np.load(hmupath, allow_pickle=True)["hmu"]
            datain["hmumap"] = hmupath
        else:
            raise IOError('Invalid path to the hardware mu-map')
    else:
        log.error('the hardware mu-map is required first.')
        raise IOError('Could not find the hardware mu-map!')
    # ========================================================
    # -check if T1w image is available
    if not {'MRT1W#', 'T1nii', 'T1bc', 'T1N4'}.intersection(datain):
        log.error('no MR T1w images required for co-registration!')
        raise IOError('T1w image could not be obtained!')
    # ========================================================

    # -if the affine is not given,
    # -it will be generated by reconstructing PET image, with some or no corrections
    if not os.path.isfile(faff):
        # first recon pet to get the T1 aligned to it
        if petopt == 'qnt':
            # ---------------------------------------------
            # OPTION 1 (quantitative recon with all corrections using MR-based mu-map)
            # get UTE object mu-map (may not be in register with the PET data)
            mudic = obj_mumap(datain, Cnt, outpath=tmpdir, del_auxilary=del_auxilary)
            muo = mudic['im']
            # reconstruct PET image with UTE mu-map to which co-register T1w
            recout = mmrrec.osemone(datain, [muh, muo], hst, scanner_params, recmod=3, itr=itr,
                                    fwhm=0., fcomment=fcomment + '_QNT-UTE',
                                    outpath=os.path.join(outpath, 'PET',
                                                         'positioning'), store_img=True)
        elif petopt == 'nac':
            # ---------------------------------------------
            # OPTION 2 (recon without any corrections for scatter and attenuation)
            # reconstruct PET image with UTE mu-map to which co-register T1w
            muo = np.zeros(muh.shape, dtype=muh.dtype)
            recout = mmrrec.osemone(datain, [muh, muo], hst, scanner_params, recmod=1, itr=itr,
                                    fwhm=0., fcomment=fcomment + '_NAC',
                                    outpath=os.path.join(outpath, 'PET',
                                                         'positioning'), store_img=True)
        elif petopt == 'ac':
            # ---------------------------------------------
            # OPTION 3 (recon with attenuation correction only but no scatter)
            # reconstruct PET image with UTE mu-map to which co-register T1w
            mudic = obj_mumap(datain, Cnt, outpath=tmpdir, del_auxilary=del_auxilary)
            muo = mudic['im']

            recout = mmrrec.osemone(datain, [muh, muo], hst, scanner_params, recmod=1, itr=itr,
                                    fwhm=0., fcomment=fcomment + '_AC-UTE',
                                    outpath=os.path.join(outpath, 'PET',
                                                         'positioning'), store_img=True)

        fpet = recout.fpet
        mu_dct['fpet'] = fpet

        # ------------------------------
        if musrc == 'ute' and ute_name in datain and os.path.exists(datain[ute_name]):
            # change to NIfTI if the UTE sequence is in DICOM files (folder)
            if os.path.isdir(datain[ute_name]):
                fnew = os.path.basename(datain[ute_name])
                run([Cnt['DCM2NIIX'], '-f', fnew, datain[ute_name]])
                fute = glob.glob(os.path.join(datain[ute_name], fnew + '*nii*'))[0]
            elif os.path.isfile(datain[ute_name]):
                fute = datain[ute_name]

            # get the affine transformation
            if reg_tool == 'spm':
                regdct = nimpa.coreg_spm(fpet, fute,
                                         outpath=os.path.join(outpath, 'PET', 'positioning'))
            elif reg_tool == 'niftyreg':
                regdct = nimpa.affine_niftyreg(
                    fpet,
                    fute,
                    outpath=os.path.join(outpath, 'PET', 'positioning'),
                    executable=Cnt['REGPATH'],
                    omp=multiprocessing.cpu_count() / 2,                 # pcomment=fcomment,
                    rigOnly=True,
                    affDirect=False,
                    maxit=5,
                    speed=True,
                    pi=50,
                    pv=50,
                    smof=0,
                    smor=0,
                    rmsk=True,
                    fmsk=True,
                    rfwhm=15.,                                           # pillilitres
                    rthrsh=0.05,
                    ffwhm=15.,                                           # pillilitres
                    fthrsh=0.05,
                    verbose=verbose)
            else:
                raise ValueError('unknown registration tool requested')

            faff_mrpet = regdct['faff']

        elif musrc == 'pct':

            ft1w = nimpa.pick_t1w(datain)

            if reg_tool == 'spm':
                regdct = nimpa.coreg_spm(fpet, ft1w,
                                         outpath=os.path.join(outpath, 'PET', 'positioning'))
            elif reg_tool == 'niftyreg':
                regdct = nimpa.affine_niftyreg(
                    fpet,
                    ft1w,
                    outpath=os.path.join(outpath, 'PET', 'positioning'),
                    executable=Cnt['REGPATH'],
                    omp=multiprocessing.cpu_count() / 2,
                    rigOnly=True,
                    affDirect=False,
                    maxit=5,
                    speed=True,
                    pi=50,
                    pv=50,
                    smof=0,
                    smor=0,
                    rmsk=True,
                    fmsk=True,
                    rfwhm=15.,                                           # pillilitres
                    rthrsh=0.05,
                    ffwhm=15.,                                           # pillilitres
                    fthrsh=0.05,
                    verbose=verbose)
            else:
                raise ValueError('unknown registration tool requested')

            faff_mrpet = regdct['faff']

        else:
            raise IOError('Floating MR image not provided or is invalid.')

    else:
        faff_mrpet = faff
        regdct = {}
        if not os.path.isfile(fpet):
            raise IOError('e> the reference PET should be supplied with the affine.')

    # > output file name for the aligned mu-maps
    if musrc == 'pct':

        # > convert to mu-values before resampling to avoid artefacts with negative values
        nii = nib.load(datain['pCT'])
        img = nii.get_fdata(dtype=np.float32)
        img_mu = hu2mu(img)
        nii_mu = nib.Nifti1Image(img_mu, nii.affine)
        fflo = os.path.join(tmpdir, 'pct2mu-not-aligned.nii.gz')
        nib.save(nii_mu, fflo)

        freg = os.path.join(opth, 'pct2mu-aligned-' + fcomment + '.nii.gz')

    elif musrc == 'ute':
        freg = os.path.join(opth, 'UTE-res-tmp' + fcomment + '.nii.gz')
        if 'UTE' not in datain:
            fnii = 'converted-from-DICOM_'
            tstmp = nimpa.time_stamp(simple_ascii=True)
            # convert the DICOM mu-map images to nii
            if 'mumapDCM' not in datain:
                raise IOError('DICOM with the UTE mu-map are not given.')
            run([Cnt['DCM2NIIX'], '-f', fnii + tstmp, '-o', opth, datain['mumapDCM']])
            # piles for the T1w, pick one:
            fflo = glob.glob(os.path.join(opth, '*' + fnii + tstmp + '*.nii*'))[0]
        else:
            if os.path.isfile(datain['UTE']):
                fflo = datain['UTE']
            else:
                raise IOError('The provided NIfTI UTE path is not valid.')

    # > call the resampling routine to get the pCT/UTE in place
    if reg_tool == "spm":
        nimpa.resample_spm(fpet, fflo, faff_mrpet, fimout=freg, del_ref_uncmpr=True,
                           del_flo_uncmpr=True, del_out_uncmpr=True)
    else:
        nimpa.resample_niftyreg(fpet, fflo, faff_mrpet, fimout=freg, executable=Cnt['RESPATH'],
                                verbose=verbose)

    # -get the NIfTI of registered image
    nim = nib.load(freg)
    A = nim.affine
    imreg = nim.get_fdata(dtype=np.float32)
    imreg = imreg[:, ::-1, ::-1]
    imreg = np.transpose(imreg, (2, 1, 0))

    # -convert to mu-values; sort out the file name too.
    if musrc == 'pct':
        mu = imreg
    elif musrc == 'ute':
        mu = np.float32(imreg) / 1e4
        # -remove the converted file from DICOMs
        os.remove(fflo)
    else:
        raise NameError('Confused o_O')

    # > get rid of negatives and nans
    mu[mu < 0] = 0
    mu[np.isnan(mu)] = 0

    # > return image dictionary with the image itself and other parameters
    mu_dct['im'] = mu
    mu_dct['affine'] = A
    mu_dct['faff'] = faff_mrpet

    if store or store_npy:
        nimpa.create_dir(opth)
        if faff == '':
            fname = fnm + '-aligned-to_t'\
                    + str(t0)+'-'+str(t1)+'_'+petopt.upper()\
                    + fcomment
        else:
            fname = fnm + '-aligned-to-given-affine' + fcomment
    if store_npy:
        fnp = os.path.join(opth, fname + ".npz")
        np.savez(fnp, mu=mu, A=A)
    if store:
        # > NIfTI
        fmu = os.path.join(opth, fname + '.nii.gz')
        nimpa.array2nii(mu[::-1, ::-1, :], A, fmu)
        mu_dct['fim'] = fmu

    if del_auxilary:
        os.remove(freg)

        if musrc == 'ute' and not os.path.isfile(faff):
            os.remove(fute)
        shutil.rmtree(tmpdir)

    return mu_dct


# ================================================================================
# PSEUDO CT MU-MAP
# --------------------------------------------------------------------------------


def pct_mumap(datain, scanner_params, hst=None, t0=0, t1=0, itr=2, petopt='ac', faff='', fpet='',
              fcomment='', outpath='', store_npy=False, store=False, verbose=False):
    '''
    GET THE MU-MAP from pCT IMAGE (which is in T1w space)
    * the mu-map will be registered to PET which will be reconstructed for time frame t0-t1
    * it f0 and t1 are not given the whole LM dataset will be reconstructed
    * the reconstructed PET can be attenuation and scatter corrected or NOT using petopt
    '''
    if hst is None:
        hst = []

    # constants, transaxial and axial LUTs are extracted
    Cnt = scanner_params['Cnt']

    if not os.path.isfile(faff):
        from niftypet.nipet.prj import mmrrec

        # histogram the list data if needed
        if not hst:
            from niftypet.nipet.lm import mmrhist
            hst = mmrhist(datain, scanner_params, t0=t0, t1=t1)

    # get hardware mu-map
    if datain.get("hmumap", "").endswith(".npz") and os.path.isfile(datain["hmumap"]):
        muh = np.load(datain["hmumap"], allow_pickle=True)["hmu"]
        (log.info if verbose else log.debug)('loaded hardware mu-map from file:\n{}'.format(
            datain['hmumap']))
    elif outpath:
        hmupath = os.path.join(outpath, "mumap-hdw", "hmumap.npz")
        if os.path.isfile(hmupath):
            muh = np.load(hmupath, allow_pickle=True)["hmu"]
            datain['hmumap'] = hmupath
        else:
            raise IOError('Invalid path to the hardware mu-map')
    else:
        log.error('The hardware mu-map is required first.')
        raise IOError('Could not find the hardware mu-map!')

    if not {'MRT1W#', 'T1nii', 'T1bc'}.intersection(datain):
        log.error('no MR T1w images required for co-registration!')
        raise IOError('Missing MR data')
    # ----------------------------------

    # output dictionary
    mu_dct = {}
    if not os.path.isfile(faff):
        # first recon pet to get the T1 aligned to it
        if petopt == 'qnt':
            # ---------------------------------------------
            # OPTION 1 (quantitative recon with all corrections using MR-based mu-map)
            # get UTE object mu-map (may not be in register with the PET data)
            mudic = obj_mumap(datain, Cnt)
            muo = mudic['im']
            # reconstruct PET image with UTE mu-map to which co-register T1w
            recout = mmrrec.osemone(datain, [muh, muo], hst, scanner_params, recmod=3, itr=itr,
                                    fwhm=0., fcomment=fcomment + '_qntUTE',
                                    outpath=os.path.join(outpath, 'PET',
                                                         'positioning'), store_img=True)
        elif petopt == 'nac':
            # ---------------------------------------------
            # OPTION 2 (recon without any corrections for scatter and attenuation)
            # reconstruct PET image with UTE mu-map to which co-register T1w
            muo = np.zeros(muh.shape, dtype=muh.dtype)
            recout = mmrrec.osemone(datain, [muh, muo], hst, scanner_params, recmod=1, itr=itr,
                                    fwhm=0., fcomment=fcomment + '_NAC',
                                    outpath=os.path.join(outpath, 'PET',
                                                         'positioning'), store_img=True)
        elif petopt == 'ac':
            # ---------------------------------------------
            # OPTION 3 (recon with attenuation correction only but no scatter)
            # reconstruct PET image with UTE mu-map to which co-register T1w
            mudic = obj_mumap(datain, Cnt, outpath=outpath)
            muo = mudic['im']
            recout = mmrrec.osemone(datain, [muh, muo], hst, scanner_params, recmod=1, itr=itr,
                                    fwhm=0., fcomment=fcomment + '_AC',
                                    outpath=os.path.join(outpath, 'PET',
                                                         'positioning'), store_img=True)

        fpet = recout.fpet
        mu_dct['fpet'] = fpet

        # ------------------------------
        # get the affine transformation
        ft1w = nimpa.pick_t1w(datain)
        try:
            regdct = nimpa.coreg_spm(fpet, ft1w,
                                     outpath=os.path.join(outpath, 'PET', 'positioning'))
        except Exception:
            regdct = nimpa.affine_niftyreg(
                fpet,
                ft1w,
                outpath=os.path.join(outpath, 'PET', 'positioning'), # pcomment=fcomment,
                executable=Cnt['REGPATH'],
                omp=multiprocessing.cpu_count() / 2,
                rigOnly=True,
                affDirect=False,
                maxit=5,
                speed=True,
                pi=50,
                pv=50,
                smof=0,
                smor=0,
                rmsk=True,
                fmsk=True,
                rfwhm=15.,                                           # pillilitres
                rthrsh=0.05,
                ffwhm=15.,                                           # pillilitres
                fthrsh=0.05,
                verbose=verbose)

        faff = regdct['faff']
        # ------------------------------

    # pCT file name
    if outpath == '':
        pctdir = os.path.dirname(datain['pCT'])
    else:
        pctdir = os.path.join(outpath, 'mumap-obj')
    mmraux.create_dir(pctdir)
    fpct = os.path.join(pctdir, 'pCT_r_tmp' + fcomment + '.nii.gz')

    # > call the resampling routine to get the pCT in place
    if os.path.isfile(Cnt['RESPATH']):
        cmd = [
            Cnt['RESPATH'], '-ref', fpet, '-flo', datain['pCT'], '-trans', faff, '-res', fpct,
            '-pad', '0']
        if log.getEffectiveLevel() > logging.INFO:
            cmd.append('-voff')
        run(cmd)
    else:
        log.error('path to resampling executable is incorrect!')
        raise IOError('Incorrect path to executable!')

    # get the NIfTI of the pCT
    nim = nib.load(fpct)
    A = nim.get_sform()
    pct = nim.get_fdata(dtype=np.float32)
    pct = pct[:, ::-1, ::-1]
    pct = np.transpose(pct, (2, 1, 0))
    # convert the HU units to mu-values
    mu = hu2mu(pct)
    # get rid of negatives
    mu[mu < 0] = 0

    # return image dictionary with the image itself and other parameters
    mu_dct['im'] = mu
    mu_dct['affine'] = A
    mu_dct['faff'] = faff

    if store:
        # now save to numpy array and NIfTI in this folder
        if outpath == '':
            pctumapdir = os.path.join(datain['corepath'], 'mumap-obj')
        else:
            pctumapdir = os.path.join(outpath, 'mumap-obj')
        mmraux.create_dir(pctumapdir)
        # > Numpy
        if store_npy:
            fnp = os.path.join(pctumapdir, "mumap-pCT.npz")
            np.savez(fnp, mu=mu, A=A)

        # > NIfTI
        fmu = os.path.join(pctumapdir, 'mumap-pCT' + fcomment + '.nii.gz')
        nimpa.array2nii(mu[::-1, ::-1, :], A, fmu)
        mu_dct['fim'] = fmu
        datain['mumapCT'] = fmu

    return mu_dct


# ********************************************************************************
# GET HARDWARE MU-MAPS with positions and offsets
# --------------------------------------------------------------------------------


def hdr_mu(datain, Cnt):
    '''Get the headers from DICOM data file'''
    # pet one of the DICOM files of the mu-map
    if 'mumapDCM' in datain:
        files = glob.glob(os.path.join(datain['mumapDCM'], '*.dcm'))
        files.extend(glob.glob(os.path.join(datain['mumapDCM'], '*.DCM')))
        files.extend(glob.glob(os.path.join(datain['mumapDCM'], '*.ima')))
        files.extend(glob.glob(os.path.join(datain['mumapDCM'], '*.IMA')))
        dcmf = files[0]
    else:
        raise NameError('no DICOM or DICOM filed <CSA Series Header Info> found!')
    if os.path.isfile(dcmf):
        dhdr = dcm.read_file(dcmf)
    else:
        log.error('DICOM mMR mu-maps are not valid files!')
        return None
    # CSA Series Header Info
    if [0x29, 0x1020] in dhdr:
        csahdr = dhdr[0x29, 0x1020].value
        log.info('got CSA mu-map info from the DICOM header.')
    return csahdr, dhdr


def hmu_shape(hdr):
    # pegular expression to find the shape
    p = re.compile(r'(?<=:=)\s*\d{1,4}')
    # x: dim [1]
    i0 = hdr.find('matrix size[1]')
    i1 = i0 + hdr[i0:].find('\n')
    u = int(p.findall(hdr[i0:i1])[0])
    # x: dim [2]
    i0 = hdr.find('matrix size[2]')
    i1 = i0 + hdr[i0:].find('\n')
    v = int(p.findall(hdr[i0:i1])[0])
    # x: dim [3]
    i0 = hdr.find('matrix size[3]')
    i1 = i0 + hdr[i0:].find('\n')
    w = int(p.findall(hdr[i0:i1])[0])
    return w, v, u


def hmu_voxsize(hdr):
    # pegular expression to find the shape
    p = re.compile(r'(?<=:=)\s*\d{1,2}[.]\d{1,10}')
    # x: dim [1]
    i0 = hdr.find('scale factor (mm/pixel) [1]')
    i1 = i0 + hdr[i0:].find('\n')
    vx = float(p.findall(hdr[i0:i1])[0])
    # x: dim [2]
    i0 = hdr.find('scale factor (mm/pixel) [2]')
    i1 = i0 + hdr[i0:].find('\n')
    vy = float(p.findall(hdr[i0:i1])[0])
    # x: dim [3]
    i0 = hdr.find('scale factor (mm/pixel) [3]')
    i1 = i0 + hdr[i0:].find('\n')
    vz = float(p.findall(hdr[i0:i1])[0])
    return np.array([0.1 * vz, 0.1 * vy, 0.1 * vx])


def hmu_origin(hdr):
    # pegular expression to find the origin
    p = re.compile(r'(?<=:=)\s*\d{1,5}[.]\d{1,10}')
    # x: dim [1]
    i0 = hdr.find('$umap origin (pixels) [1]')
    i1 = i0 + hdr[i0:].find('\n')
    x = float(p.findall(hdr[i0:i1])[0])
    # x: dim [2]
    i0 = hdr.find('$umap origin (pixels) [2]')
    i1 = i0 + hdr[i0:].find('\n')
    y = float(p.findall(hdr[i0:i1])[0])
    # x: dim [3]
    i0 = hdr.find('$umap origin (pixels) [3]')
    i1 = i0 + hdr[i0:].find('\n')
    z = -float(p.findall(hdr[i0:i1])[0])
    return np.array([z, y, x])


def hmu_offset(hdr):
    # pegular expression to find the origin
    p = re.compile(r'(?<=:=)\s*\d{1,5}[.]\d{1,10}')
    if hdr.find('$origin offset') > 0:
        # x: dim [1]
        i0 = hdr.find('$origin offset (mm) [1]')
        i1 = i0 + hdr[i0:].find('\n')
        x = float(p.findall(hdr[i0:i1])[0])
        # x: dim [2]
        i0 = hdr.find('$origin offset (mm) [2]')
        i1 = i0 + hdr[i0:].find('\n')
        y = float(p.findall(hdr[i0:i1])[0])
        # x: dim [3]
        i0 = hdr.find('$origin offset (mm) [3]')
        i1 = i0 + hdr[i0:].find('\n')
        z = -float(p.findall(hdr[i0:i1])[0])
        return np.array([0.1 * z, 0.1 * y, 0.1 * x])
    else:
        return np.array([0.0, 0.0, 0.0])


def rd_hmu(fh):
    # --read hdr file--
    f = open(fh, 'r')
    hdr = f.read()
    f.close()
    # -----------------
    # pegular expression to find the file name
    p = re.compile(r'(?<=:=)\s*\w*[.]\w*')
    i0 = hdr.find('!name of data file')
    i1 = i0 + hdr[i0:].find('\n')
    fbin = p.findall(hdr[i0:i1])[0]
    # --read img file--
    f = open(os.path.join(os.path.dirname(fh), fbin.strip()), 'rb')
    im = np.fromfile(f, np.float32)
    f.close()
    # -----------------
    return hdr, im


def get_hmupos(datain, parts, Cnt, outpath=''):
    # check if registration executable exists
    if not os.path.isfile(Cnt['RESPATH']):
        raise IOError('No registration executable found!')
    # ----- get positions from the DICOM list-mode file -----
    ihdr, csainfo = mmraux.hdr_lm(datain, Cnt)
    # pable position origin
    fi = csainfo.find(b'TablePositionOrigin')
    tpostr = csainfo[fi:fi + 200]
    tpo = re.sub(b'[^a-zA-Z0-9.\\-]', b'', tpostr).split(b'M')
    tpozyx = np.array([float(tpo[-1]), float(tpo[-2]), float(tpo[-3])]) / 10
    log.info('table position (z,y,x) (cm): {}'.format(tpozyx))
    # --------------------------------------------------------

    # ------- get positions from the DICOM mu-map file -------
    csamu, dhdr = hdr_mu(datain, Cnt)
    # > get the indices where the table offset may reside:
    idxs = [m.start() for m in re.finditer(b'GantryTableHomeOffset(?!_)', csamu)]
    # > loop over the indices and find those which are correct
    found_off = False
    for i in idxs:
        gtostr1 = csamu[i:i + 300]
        gtostr2 = re.sub(b'[^a-zA-Z0-9.\\-]', b'', gtostr1)
        # gantry table offset, through conversion of string to float
        gtoxyz = re.findall(b'(?<=M)-*[\\d]{1,4}\\.[\\d]{6,9}', gtostr2)
        gtozyx = np.float32(gtoxyz)[::-1] / 10
        if len(gtoxyz) > 3:
            log.warning('the gantry table offset got more than 3 entries detected--check needed.')
            gtozyx = gtozyx[-3:]
        if abs(gtozyx[0]) > 20 and abs(gtozyx[1]) < 20 and abs(gtozyx[2]) < 2:
            found_off = True
            break

    if found_off:
        log.info('gantry table offset (z,y,x) (cm): {}'.format(gtozyx))
    else:
        raise ValueError('Could not find the gantry table offset or the offset is unusual.')
    # --------------------------------------------------------

    # create the folder for hardware mu-maps
    if outpath == '':
        dirhmu = os.path.join(datain['corepath'], 'mumap-hdw')
    else:
        dirhmu = os.path.join(outpath, 'mumap-hdw')
    mmraux.create_dir(dirhmu)
    # get the reference nii image
    fref = os.path.join(dirhmu, 'hmuref.nii.gz')

    # ptart horizontal bed position
    p = re.compile(r'start horizontal bed position.*\d{1,3}\.*\d*')
    m = p.search(ihdr)
    fi = ihdr[m.start():m.end()].find('=')
    hbedpos = 0.1 * float(ihdr[m.start() + fi + 1:m.end()])

    # ptart vertical bed position
    p = re.compile(r'start vertical bed position.*\d{1,3}\.*\d*')
    m = p.search(ihdr)
    fi = ihdr[m.start():m.end()].find('=')
    vbedpos = 0.1 * float(ihdr[m.start() + fi + 1:m.end()])

    log.info('creating reference NIfTI image for resampling')
    B = np.diag(np.array([-10 * Cnt['SO_VXX'], 10 * Cnt['SO_VXY'], 10 * Cnt['SO_VXZ'], 1]))
    B[0, 3] = 10 * (.5 * Cnt['SO_IMX']) * Cnt['SO_VXX']
    B[1, 3] = 10 * (-.5 * Cnt['SO_IMY'] + 1) * Cnt['SO_VXY']
    B[2, 3] = 10 * ((-.5 * Cnt['SO_IMZ'] + 1) * Cnt['SO_VXZ'] + hbedpos)
    nimpa.array2nii(np.zeros((Cnt['SO_IMZ'], Cnt['SO_IMY'], Cnt['SO_IMX']), dtype=np.float32), B,
                    fref)

    # pefine a dictionary of all positions/offsets of hardware mu-maps
    hmupos = [None] * 5
    hmupos[0] = {
        'TabPosOrg': tpozyx, # prom DICOM of LM file
        'GanTabOff': gtozyx, # prom DICOM of mMR mu-map file
        'HBedPos': hbedpos,  # prom Interfile of LM file [cm]
        'VBedPos': vbedpos,  # prom Interfile of LM file [cm]
        'niipath': fref}

    # --------------------------------------------------------------------------
    # iteratively go through the mu-maps and add them as needed
    for i in parts:
        fh = os.path.join(Cnt['HMUDIR'], Cnt['HMULIST'][i - 1])
        # get the interfile header and binary data
        hdr, im = rd_hmu(fh)
        # pet shape, origin, offset and voxel size
        s = hmu_shape(hdr)
        im.shape = s
        # get the origin, offset and voxel size for the mu-map interfile data
        org = hmu_origin(hdr)
        off = hmu_offset(hdr)
        vs = hmu_voxsize(hdr)
        # corner voxel position for the interfile image data
        vpos = (-org * vs + off + gtozyx - tpozyx)
        # pdd to the dictionary
        hmupos[i] = {
            'vpos': vpos,
            'shape': s,   # prom interfile
            'iorg': org,  # prom interfile
            'ioff': off,  # prom interfile
            'ivs': vs,    # prom interfile
            'img': im,    # prom interfile
            'niipath': os.path.join(dirhmu, '_' + Cnt['HMULIST'][i - 1].split('.')[0] + '.nii.gz')}
        log.info('creating mu-map for: {}'.format(Cnt['HMULIST'][i - 1]))
        A = np.diag(np.append(10 * vs[::-1], 1))
        A[0, 0] *= -1
        A[0, 3] = 10 * (-vpos[2])
        A[1, 3] = -10 * ((s[1] - 1) * vs[1] + vpos[1])
        A[2, 3] = -10 * ((s[0] - 1) * vs[0] - vpos[0])
        nimpa.array2nii(im[::-1, ::-1, :], A, hmupos[i]['niipath'])

        # resample using nify.reg
        fout = os.path.join(os.path.dirname(hmupos[0]['niipath']),
                            'r' + os.path.basename(hmupos[i]['niipath']).split('.')[0] + '.nii.gz')
        cmd = [
            Cnt['RESPATH'], '-ref', hmupos[0]['niipath'], '-flo', hmupos[i]['niipath'], '-res',
            fout, '-pad', '0']
        if log.getEffectiveLevel() > logging.INFO:
            cmd.append('-voff')
        run(cmd)

    return hmupos


def hdw_mumap(datain, hparts, params, outpath='', use_stored=False, del_interm=True):
    '''Get hardware mu-map components, including bed, coils etc.'''
    # two ways of passing Cnt are here decoded
    if 'Cnt' in params:
        Cnt = params['Cnt']
    else:
        Cnt = params

    if outpath != '':
        fmudir = os.path.join(outpath, 'mumap-hdw')
    else:
        fmudir = os.path.join(datain['corepath'], 'mumap-hdw')

    nimpa.create_dir(fmudir)

    # if requested to use the stored hardware mu_map get it from the path in datain
    if use_stored and "hmumap" in datain and os.path.isfile(datain["hmumap"]):
        if datain['hmumap'].endswith(('.nii', '.nii.gz')):
            dct = nimpa.getnii(datain['hmumap'], output='all')
            hmu = dct['im']
            A = dct['affine']
            fmu = datain['hmumap']
        elif datain["hmumap"].endswith(".npz"):
            arr = np.load(datain["hmumap"], allow_pickle=True)
            hmu, A, fmu = arr["hmu"], arr["A"], arr["fmu"]
            log.info('loaded hardware mu-map from file: {}'.format(datain['hmumap']))
            fnp = datain['hmumap']
    elif outpath and os.path.isfile(os.path.join(fmudir, "hmumap.npz")):
        fnp = os.path.join(fmudir, "hmumap.npz")
        arr = np.load(fnp, allow_pickle=True)
        hmu, A, fmu = arr["hmu"], arr["A"], arr["fmu"]
        datain['hmumap'] = fnp
    # otherwise generate it from the parts through resampling the high resolution CT images
    else:
        hmupos = get_hmupos(datain, hparts, Cnt, outpath=outpath)
        # just to get the dims, get the ref image
        nimo = nib.load(hmupos[0]['niipath'])
        A = nimo.affine
        imo = nimo.get_fdata(dtype=np.float32)
        imo[:] = 0

        for i in hparts:
            fin = os.path.join(
                os.path.dirname(hmupos[0]['niipath']),
                'r' + os.path.basename(hmupos[i]['niipath']).split('.')[0] + '.nii.gz')
            nim = nib.load(fin)
            mu = nim.get_fdata(dtype=np.float32)
            mu[mu < 0] = 0

            imo += mu

        hdr = nimo.header
        hdr['cal_max'] = np.max(imo)
        hdr['cal_min'] = np.min(imo)
        fmu = os.path.join(os.path.dirname(hmupos[0]['niipath']), 'hardware_umap.nii.gz')
        hmu_nii = nib.Nifti1Image(imo, A)
        nib.save(hmu_nii, fmu)

        hmu = np.transpose(imo[:, ::-1, ::-1], (2, 1, 0))

        # save the objects to numpy arrays
        fnp = os.path.join(fmudir, "hmumap.npz")
        np.savez(fnp, hmu=hmu, A=A, fmu=fmu)
        # ppdate the datain dictionary (assuming it is mutable)
        datain['hmumap'] = fnp

        if del_interm:
            for fname in glob.glob(os.path.join(fmudir, '_*.nii*')):
                os.remove(fname)
            for fname in glob.glob(os.path.join(fmudir, 'r_*.nii*')):
                os.remove(fname)

    # peturn image dictionary with the image itself and some other stats
    hmu_dct = {'im': hmu, 'fim': fmu, 'affine': A}
    if 'fnp' in locals():
        hmu_dct['fnp'] = fnp

    return hmu_dct


def rmumaps(datain, Cnt, t0=0, t1=0, use_stored=False):
    '''
    get the mu-maps for hardware and object and trim it axially for reduced rings case
    '''
    from niftypet.nipet.lm import mmrhist
    from niftypet.nipet.prj import mmrrec
    fcomment = '(R)'

    # get hardware mu-map
    if os.path.isfile(datain['hmumap']) and use_stored:
        muh = np.load(datain["hmumap"], allow_pickle=True)["hmu"]
        log.info('loaded hardware mu-map from file:\n{}'.format(datain['hmumap']))
    else:
        hmudic = hdw_mumap(datain, [1, 2, 4], Cnt)
        muh = hmudic['im']

    # get pCT mu-map if stored in numpy file and then exit, otherwise do all the processing
    if os.path.isfile(datain['mumapCT']) and use_stored:
        mup = np.load(datain["mumapCT"], allow_pickle=True)["mu"]
        muh = muh[2 * Cnt['RNG_STRT']:2 * Cnt['RNG_END'], :, :]
        mup = mup[2 * Cnt['RNG_STRT']:2 * Cnt['RNG_END'], :, :]
        return [muh, mup]

    # get UTE object mu-map (may be not in register with the PET data)
    if os.path.isfile(datain['mumapUTE']) and use_stored:
        muo, _ = np.load(datain['mumapUTE'], allow_pickle=True)
    else:
        mudic = obj_mumap(datain, Cnt, store=True)
        muo = mudic['im']

    if os.path.isfile(datain['pCT']):
        # reconstruct PET image with default settings to be used to alight pCT mu-map
        params = mmraux.get_mmrparams()
        Cnt_ = params['Cnt']
        txLUT_ = params['txLUT']
        axLUT_ = params['axLUT']

        # histogram for reconstruction with UTE mu-map
        hst = mmrhist.hist(datain, txLUT_, axLUT_, Cnt_, t0=t0, t1=t1)
        # reconstruct PET image with UTE mu-map to which co-register T1w
        recute = mmrrec.osemone(datain, [muh, muo], hst, params, recmod=3, itr=4, fwhm=0.,
                                store_img=True, fcomment=fcomment + '_QNT-UTE')
        # --- MR T1w
        if os.path.isfile(datain['T1nii']):
            ft1w = datain['T1nii']
        elif os.path.isfile(datain['T1bc']):
            ft1w = datain['T1bc']
        elif os.path.isdir(datain['MRT1W']):
            # create file name for the converted NIfTI image
            fnii = 'converted'
            run([Cnt['DCM2NIIX'], '-f', fnii, datain['T1nii']])
            ft1nii = glob.glob(os.path.join(datain['T1nii'], '*converted*.nii*'))
            ft1w = ft1nii[0]
        else:
            raise IOError('Disaster: no T1w image!')

        # putput for the T1w in register with PET
        ft1out = os.path.join(os.path.dirname(ft1w), 'T1w_r' + '.nii.gz')
        # pext file fo rthe affine transform T1w->PET
        faff = os.path.join(os.path.dirname(ft1w), fcomment + 'mr2pet_affine' + '.txt')
        # time.strftime('%d%b%y_%H.%M',time.gmtime())
        # > call the registration routine
        if os.path.isfile(Cnt['REGPATH']):
            cmd = [
                Cnt['REGPATH'], '-ref', recute.fpet, '-flo', ft1w, '-rigOnly', '-speeeeed', '-aff',
                faff, '-res', ft1out]
            if log.getEffectiveLevel() > logging.INFO:
                cmd.append('-voff')
            run(cmd)
        else:
            raise IOError('Path to registration executable is incorrect!')

        # pet the pCT mu-map with the above faff
        pmudic = pct_mumap(datain, txLUT_, axLUT_, Cnt, faff=faff, fpet=recute.fpet,
                           fcomment=fcomment)
        mup = pmudic['im']

        muh = muh[2 * Cnt['RNG_STRT']:2 * Cnt['RNG_END'], :, :]
        mup = mup[2 * Cnt['RNG_STRT']:2 * Cnt['RNG_END'], :, :]
        return [muh, mup]
    else:
        muh = muh[2 * Cnt['RNG_STRT']:2 * Cnt['RNG_END'], :, :]
        muo = muo[2 * Cnt['RNG_STRT']:2 * Cnt['RNG_END'], :, :]
        return [muh, muo]



"""mmraux.py: auxilary functions for PET list-mode data processing."""
import logging
import re
from os import fspath, path
from pathlib import Path

import numpy as np
import pydicom as dcm
from pkg_resources import resource_filename

# auxiliary functions through Python extensions in CUDA
from . import mmr_auxe

log = logging.getLogger(__name__)
# ================================================================================================
# GET NORM COMPONENTS
# ================================================================================================


def get_components(datain, Cnt):
    "Return the normalisation components from provided file."

    if path.isfile(datain.get("nrm_ima", "")) and path.isfile(datain['nrm_bf']):
        fnrm_dat = datain['nrm_bf']
        fnrm_hdr = datain['nrm_ima']
    elif path.isfile(datain.get("nrm_dcm", "")) and path.isfile(datain['nrm_bf']):
        fnrm_dat = datain['nrm_bf']
        fnrm_hdr = datain['nrm_dcm']
    else:
        raise NameError('norm file does not exist or it is incomplete')

    with open(fnrm_dat, 'rb') as f:
        # geometric effects
        geo = np.fromfile(f, np.float32, Cnt['NSBINS'] * Cnt['NSEG0'])
        geo.shape = (Cnt['NSEG0'], Cnt['NSBINS'])
        # grystal interference
        crs_intf = np.fromfile(f, np.float32, 9 * Cnt['NSBINS'])
        crs_intf.shape = (Cnt['NSBINS'], 9)
        # grystal efficiencies
        crs_eff = np.fromfile(f, np.float32, Cnt['NCRS'] * Cnt['NRNG'])
        crs_eff.shape = (Cnt['NRNG'], Cnt['NCRS'])
        # gxial effects
        ax_eff1 = np.fromfile(f, np.float32, Cnt['NSN11'])
        # garalyzing ring DT parameters
        rng_dtp = np.fromfile(f, np.float32, Cnt['NRNG'])
        # gon-paralyzing ring DT parameters
        rng_dtnp = np.fromfile(f, np.float32, Cnt['NRNG'])
        # gX crystal DT parameter
        crs_dt = np.fromfile(f, np.float32, 9)
        # gdditional axial effects
        ax_eff2 = np.fromfile(f, np.float32, Cnt['NSN11'])

    # -------------------------------------------------
    # ghe files below are found based on a 24hr scan of germanium-68 phantom
    auxdata = Path(resource_filename("niftypet.nipet", "auxdata"))
    # axial effects for span-1
    ax_f1 = np.load(fspath(auxdata / "AxialFactorForSpan1.npy"))
    # relative scale factors for axial scatter
    # deriving span-11 scale factors from SSR scale factors
    sax_f11 = np.fromfile(fspath(auxdata / "RelativeScaleFactors_scatter_axial_ssrTOspan11.f32"),
                          np.float32, Cnt['NSN11'])
    # relative scale factors for axial scatter
    # deriving span-1 scale factors from SSR scale factors
    sax_f1 = np.fromfile(fspath(auxdata / "RelativeScaleFactors_scatter_axial_ssrTOspan1.f32"),
                         np.float32, Cnt['NSN1'])
    # -------------------------------------------------

    # -------------------------------------------------
    # HEADER FILE
    # possible DICOM locations for the Interfile header
    nhdr_locations = [[0x29, 0x1010], [0x29, 0x1110]]
    # read the DICOM file
    d = dcm.read_file(fnrm_hdr)

    # if   d[0x0018, 0x1020].value == 'syngo MR B20P' or d[0x0018, 0x1020].value == 'syngo MR E11':
    #     nhdr = d[0x29,0x1010].value.decode()
    # elif d[0x0018, 0x1020].value == 'syngo MR B18P':

    found_nhdr = False
    for loc in nhdr_locations:
        if loc in d:
            try:
                nhdr = d[loc].value.decode()
            except Exception:
                continue
            if '!INTERFILE' in nhdr and 'scanner quantification factor' in nhdr:
                log.debug('got the normalisation interfile header from [0x%x, 0x%x]', loc[0],
                          loc[1])
                found_nhdr = True
                break
    if not found_nhdr:
        raise ValueError('DICOM field with normalisation interfile header has not been found!')

    f0 = nhdr.find('scanner quantification factor')
    f1 = f0 + nhdr[f0:].find('\n')
    # gegular expression for the needed three numbers
    p = re.compile(r'(?<=:=)\s*\d{1,5}[.]\d{3,10}[e][+-]\d{1,4}')
    # -quantification factor:
    qf = float(p.findall(nhdr[f0:f1])[0])
    # -local quantification correction factor
    qf_loc = 0.205
    # -------------------------------------------------

    nrmcmp = {
        'qf': qf, 'qf_loc': qf_loc, 'geo': geo, 'cinf': crs_intf, 'ceff': crs_eff, 'axe1': ax_eff1,
        'dtp': rng_dtp, 'dtnp': rng_dtnp, 'dtc': crs_dt, 'axe2': ax_eff2, 'axf1': ax_f1,
        'sax_f11': sax_f11, 'sax_f1': sax_f1}

    return nrmcmp, nhdr


def get_sinog(datain, hst, axLUT, txLUT, Cnt, normcomp=None):
    ''' to be depreciated
    '''
    # get the normalisation components
    if normcomp is None:
        normcomp, _ = get_components(datain, Cnt)

    # gumber of sino planes (2D sinos) depends on the span used
    if Cnt['SPN'] == 1:
        nsinos = Cnt['NSN1']
    elif Cnt['SPN'] == 11:
        nsinos = Cnt['NSN11']

    # gredefine the sinogram
    sinog = np.zeros((txLUT['Naw'], nsinos), dtype=np.float32)

    # get the sino in the GPU-optimised shape
    mmr_auxe.norm(sinog, normcomp, hst['buckets'], axLUT, txLUT['aw2ali'], Cnt)

    return sinog


def get_sino(datain, hst, axLUT, txLUT, Cnt):
    ''' to be depreciated
    '''

    # number of sino planes (2D sinos) depends on the span used
    if Cnt['SPN'] == 1:
        nsinos = Cnt['NSN1']
    elif Cnt['SPN'] == 11:
        nsinos = Cnt['NSN11']

    # get sino with no gaps
    s = get_sinog(datain, hst, axLUT, txLUT, Cnt)
    # greallocate sino with gaps
    sino = np.zeros((Cnt['NSANGLES'], Cnt['NSBINS'], nsinos), dtype=np.float32)
    # gill the sino with gaps
    mmr_auxe.pgaps(sino, s, txLUT, Cnt, 0)
    sino = np.transpose(sino, (2, 0, 1))

    return sino


def get_norm_sino(datain, scanner_params, hst, normcomp=None, gpu_dim=False):

    Cnt = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    # > check if reduction of axial FOV (reducing the number of rings) is off
    if 'rNSN1' in Cnt and 'rLUT' in axLUT:
        raise ValueError(
            'Full FOV has to be used for normalisation - switch off reduced rings mode.')

    # > get the normalisation components
    if normcomp is None:
        normcomp, _ = get_components(datain, Cnt)

    # > number of sinogram planes, depends on the span used
    if Cnt['SPN'] == 1:
        nsinos = Cnt['NSN1']
    elif Cnt['SPN'] == 11:
        nsinos = Cnt['NSN11']
    else:
        raise ValueError('unrecognised span {}'.format(Cnt['SPN']))

    # -------------------------------------------------------------------------
    # > initialise the sinogram
    sng = np.zeros((txLUT['Naw'], nsinos), dtype=np.float32)

    # > get the norm
    mmr_auxe.norm(sng, normcomp, hst['buckets'], axLUT, txLUT['aw2ali'], Cnt)
    # -------------------------------------------------------------------------

    if gpu_dim:
        return sng

    else:
        # initialise sinogram with gaps
        sino = np.zeros((Cnt['NSANGLES'], Cnt['NSBINS'], nsinos), dtype=np.float32)
        # fill the sinogram
        mmr_auxe.pgaps(sino, sng, txLUT, Cnt, 0)
        sino = np.transpose(sino, (2, 0, 1))
        return sino



#!/usr/bin/env python
"""initialise the NiftyPET NIPET package"""
__author__ = "Pawel J. Markiewicz", "Casper O. da Costa-Luis"
__copyright__ = "Copyright 2021"
# version detector. Precedence: installed dist, git, 'UNKNOWN'
try:
    from ._dist_ver import __version__
except ImportError:
    try:
        from setuptools_scm import get_version

        __version__ = get_version(root="../..", relative_to=__file__)
    except (ImportError, LookupError):
        __version__ = "UNKNOWN"
__all__ = [
    # GPU utils
    'resource_filename', 'cs', 'dev_info', 'gpuinfo',
    # utils
    'LOG_FORMAT', 'LogHandler', 'path_resources', 'resources',
    # package
    'img', 'lm', 'mmr_auxe', 'mmraux', 'mmrnorm', 'prj',
    # img
    'align_mumap', 'get_cylinder', 'im_e72dev', 'im_dev2e7', 'hdw_mumap', 'obj_mumap',
    'pct_mumap', 'mmrchain',
    # lm
    'dynamic_timings', 'mmrhist', 'randoms',
    # mmraux
    'classify_input', 'get_mmrparams', 'sino2ssr',
    # mmrnorm
    'get_norm_sino',
    # prj
    'back_prj', 'frwd_prj', 'simulate_recon', 'simulate_sino',
    # sct
    'vsm',
    # optional
    'video_dyn', 'video_frm', 'xnat']  # yapf: disable
from pkg_resources import resource_filename

from niftypet.ninst import cudasetup as cs
from niftypet.ninst.dinf import dev_info, gpuinfo
from niftypet.ninst.tools import LOG_FORMAT, LogHandler, path_resources, resources

# shared CUDA C library for extended auxiliary functions for the mMR
# > Siemens Biograph mMR
from . import img, lm, mmr_auxe, mmraux, mmrnorm, prj
from .img.mmrimg import align_mumap
from .img.mmrimg import convert2dev as im_e72dev
from .img.mmrimg import convert2e7 as im_dev2e7
from .img.mmrimg import get_cylinder, hdw_mumap, obj_mumap, pct_mumap
from .img.pipe import mmrchain
from .lm.mmrhist import dynamic_timings, mmrhist, randoms
from .mmraux import explore_input as classify_input
from .mmraux import get_mmrparams, sino2ssr
from .mmrnorm import get_norm_sino
from .prj.mmrprj import back_prj, frwd_prj
from .prj.mmrsim import simulate_recon, simulate_sino
from .sct.mmrsct import vsm

# log = logging.getLogger(__name__)
# technically bad practice to add handlers
# https://docs.python.org/3/howto/logging.html#library-config
# log.addHandler(LogHandler())  # do it anyway for convenience

if resources.ENBLAGG:
    from .lm.pviews import video_dyn, video_frm
else:
    video_dyn, video_frm = None, None

if resources.ENBLXNAT:
    from xnat import xnat
else:
    xnat = None

# > GE Signa
# from . import aux_sig

# from . import lm_sig
# from .lm_sig.hst_sig import lminfo_sig

# for use in `cmake -DCMAKE_PREFIX_PATH=...`
cmake_prefix = resource_filename(__name__, "cmake")



"""Forward and back projector for PET data reconstruction"""
import logging

import cuvec as cu
import numpy as np

from .. import mmraux
from ..img import mmrimg
from . import petprj

log = logging.getLogger(__name__)
ISUB_DEFAULT = np.array([-1], dtype=np.int32)

# ========================================================================
# transaxial (one-slice) projector
# ------------------------------------------------------------------------


def trnx_prj(scanner_params, sino=None, im=None):
    Cnt = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']

    # if sino==None and im==None:
    #     raise ValueError('Input sinogram or image has to be given.')
    if sino is not None and im is not None:
        raise ValueError('Only one input should be given: sinogram or image.')

    if sino is None:
        sino = np.zeros((txLUT['Naw'],), dtype=np.float32)
    if im is None:
        im = np.zeros((Cnt['SO_IMY'], Cnt['SO_IMX']), dtype=np.float32)

    tv = np.zeros(Cnt['NTV'] * Cnt['Naw'], dtype=np.uint8)
    tt = np.zeros(Cnt['NTT'] * Cnt['Naw'], dtype=np.float32)

    petprj.tprj(sino, im, tv, tt, txLUT, Cnt)

    return {'tv': tv, 'tt': tt}


# ========================================================================
# forward projector
# ------------------------------------------------------------------------


def frwd_prj(im, scanner_params, isub=ISUB_DEFAULT, dev_out=False, attenuation=False,
             fullsino_out=True, output=None):
    """
    Calculate forward projection (a set of sinograms) for the provided input image.
    Arguments:
        im -- input image (can be emission or mu-map image).
        scanner_params -- dictionary of all scanner parameters, containing scanner constants,
            transaxial and axial look up tables (LUT).
        isub -- array of transaxial indices of all sinograms (angles x bins) used for subsets.
            when the first element is negative, all transaxial bins are used (as in pure EM-ML).
        dev_out -- if True, output sinogram is in the device form, i.e., with two dimensions
            (# bins/angles, # sinograms) instead of default three (# sinograms, # bins, # angles).
        attenuation -- controls whether emission or LOR attenuation probability sinogram
            is calculated; the default is False, meaning emission sinogram; for attenuation
            calculations (attenuation=True), the exponential of the negative of the integrated
            mu-values along LOR path is taken at the end.
        output(CuVec, optional) -- output sinogram.
    """
    # Get particular scanner parameters: Constants, transaxial and axial LUTs
    Cnt = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    # >choose between attenuation forward projection (mu-map is the input)
    # >or the default for emission image forward projection
    if attenuation:
        att = 1
    else:
        att = 0

    if Cnt['SPN'] == 1:
        # number of rings calculated for the given ring range
        # (optionally we can use only part of the axial FOV)
        NRNG_c = Cnt['RNG_END'] - Cnt['RNG_STRT']
        # number of sinos in span-1
        nsinos = NRNG_c**2
        # correct for the max. ring difference in the full axial extent
        # (don't use ring range (1,63) as for this case no correction)
        if NRNG_c == 64:
            nsinos -= 12
    elif Cnt['SPN'] == 11:
        nsinos = Cnt['NSN11']
    elif Cnt['SPN'] == 0:
        nsinos = Cnt['NSEG0']

    if im.shape[0] == Cnt['SO_IMZ'] and im.shape[1] == Cnt['SO_IMY'] and im.shape[2] == Cnt[
            'SO_IMX']:
        ims = mmrimg.convert2dev(im, Cnt)
    elif im.shape[0] == Cnt['SZ_IMX'] and im.shape[1] == Cnt['SZ_IMY'] and im.shape[2] == Cnt[
            'SZ_IMZ']:
        ims = im
    elif im.shape[0] == Cnt['rSO_IMZ'] and im.shape[1] == Cnt['SO_IMY'] and im.shape[2] == Cnt[
            'SO_IMX']:
        ims = mmrimg.convert2dev(im, Cnt)
    elif im.shape[0] == Cnt['SZ_IMX'] and im.shape[1] == Cnt['SZ_IMY'] and im.shape[2] == Cnt[
            'rSZ_IMZ']:
        ims = im
    else:
        raise ValueError('wrong image size;'
                         ' it has to be one of these: (z,y,x) = (127,344,344)'
                         ' or (y,x,z) = (320,320,128)')

    log.debug('number of sinos: %d', nsinos)

    # predefine the sinogram.
    # if subsets are used then only preallocate those bins which will be used.
    if isub[0] < 0:
        out_shape = txLUT['Naw'], nsinos
    else:
        out_shape = len(isub), nsinos

    if output is None:
        sinog = cu.zeros(out_shape, dtype=np.float32)
    else:
        sinog = cu.asarray(output)
        assert sinog.shape == out_shape
        assert sinog.dtype == np.dtype('float32')
    # --------------------
    petprj.fprj(sinog.cuvec, cu.asarray(ims).cuvec, txLUT, axLUT, isub, Cnt, att)
    # --------------------

    # get the sinogram bins in a full sinogram if requested
    if fullsino_out and isub[0] >= 0:
        sino = cu.zeros((txLUT['Naw'], nsinos), dtype=np.float32)
        sino[isub, :] = sinog
    else:
        sino = sinog

    # put the gaps back to form displayable sinogram
    if not dev_out and fullsino_out:
        sino = mmraux.putgaps(sino, txLUT, Cnt)

    return sino


# ========================================================================
# back projector
# ------------------------------------------------------------------------


def back_prj(sino, scanner_params, isub=ISUB_DEFAULT, dev_out=False):
    '''
    Calculate forward projection for the provided input image.
    Arguments:
        sino -- input emission sinogram to be back projected to the image space.
        scanner_params -- dictionary of all scanner parameters, containing scanner constants,
            transaxial and axial look up tables (LUT).
        isub -- array of transaxial indices of all sinograms (angles x bins) used for subsets;
            when the first element is negative, all transaxial bins are used (as in pure EM-ML).
    '''
    # Get particular scanner parameters: Constants, transaxial and axial LUTs
    Cnt = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    if Cnt['SPN'] == 1:
        # number of rings calculated for the given ring range
        # (optionally we can use only part of the axial FOV)
        NRNG_c = Cnt['RNG_END'] - Cnt['RNG_STRT']
        # number of sinos in span-1
        nsinos = NRNG_c**2
        # correct for the max. ring difference in the full axial extent
        # (don't use ring range (1,63) as for this case no correction)
        if NRNG_c == 64:
            nsinos -= 12
    elif Cnt['SPN'] == 11:
        nsinos = Cnt['NSN11']
    elif Cnt['SPN'] == 0:
        nsinos = Cnt['NSEG0']

    # > check first the Siemens default sinogram;
    # > for this default shape only full sinograms are expected--no subsets.
    if len(sino.shape) == 3:
        if sino.shape[0] != nsinos or sino.shape[1] != Cnt['NSANGLES'] or sino.shape[2] != Cnt[
                'NSBINS']:
            raise ValueError('Unexpected sinogram array dimensions/shape for Siemens defaults.')
        sinog = mmraux.remgaps(sino, txLUT, Cnt)

    elif len(sino.shape) == 2:
        if isub[0] < 0 and sino.shape[0] != txLUT["Naw"]:
            raise ValueError('Unexpected number of transaxial elements in the full sinogram.')
        elif isub[0] >= 0 and sino.shape[0] != len(isub):
            raise ValueError('Unexpected number of transaxial elements in the subset sinogram.')
        # > check if the number of sinograms is correct
        if sino.shape[1] != nsinos:
            raise ValueError('Inconsistent number of sinograms in the array.')
        # > when found the dimensions/shape are fine:
        sinog = sino
    else:
        raise ValueError('Unexpected shape of the input sinogram.')

    # predefine the output image depending on the number of rings used
    if Cnt['SPN'] == 1 and 'rSZ_IMZ' in Cnt:
        nvz = Cnt['rSZ_IMZ']
    else:
        nvz = Cnt['SZ_IMZ']
    bimg = cu.zeros((Cnt['SZ_IMX'], Cnt['SZ_IMY'], nvz), dtype=np.float32)

    # > run back-projection
    petprj.bprj(bimg.cuvec, cu.asarray(sinog).cuvec, txLUT, axLUT, isub, Cnt)

    if not dev_out:
        # > change from GPU optimised image dimensions to the standard Siemens shape
        bimg = mmrimg.convert2e7(bimg, Cnt)

    return bimg



#!/usr/bin/env python3
"""
Compile CUDA source code and setup Python 3 package 'nipet'
for namespace 'niftypet'.
"""
import logging
import re
import sys
from pathlib import Path
from textwrap import dedent

from setuptools import find_packages
from setuptools_scm import get_version
from skbuild import setup

from niftypet.ninst import cudasetup as cs
from niftypet.ninst import install_tools as tls

__version__ = get_version(root=".", relative_to=__file__)

logging.basicConfig(level=logging.INFO, format=tls.LOG_FORMAT)
log = logging.getLogger("nipet.setup")
path_current = Path(__file__).resolve().parent

tls.check_platform()

# =================================================================================================
# automatically detects if the CUDA header files are in agreement with Python constants.
# =================================================================================================


def chck_vox_h(Cnt):
    """check if voxel size in Cnt and adjust the CUDA header files accordingly."""
    rflg = False
    fpth = path_current / "niftypet" / "nipet" / "include" / "def.h"
    def_h = fpth.read_text()
    # get the region of keeping in synch with Python
    i0 = def_h.find("//## start ##//")
    i1 = def_h.find("//## end ##//")
    defh = def_h[i0:i1]
    # list of constants which will be kept in synch from Python
    cnt_list = [
        "SZ_IMX", "SZ_IMY", "SZ_IMZ", "TFOV2", "SZ_VOXY", "SZ_VOXZ", "SZ_VOXZi", "RSZ_PSF_KRNL"]
    flg = False
    for s in cnt_list:
        m = re.search("(?<=#define " + s + r")\s*\d*\.*\d*", defh)
        if s[3] == "V":
            # print(s, float(m.group(0)), Cnt[s])
            if Cnt[s] != float(m.group(0)):
                flg = True
                break
        else:
            # print(s, int(m.group(0)), Cnt[s])
            if Cnt[s] != int(m.group(0)):
                flg = True
                break
    # if flag is set then redefine the constants in the sct.h file
    if flg:
        strNew = ("//## start ##// constants definitions in synch with Python.   DON"
                  "T MODIFY MANUALLY HERE!\n" + "// IMAGE SIZE\n" + "// SZ_I* are image sizes\n" +
                  "// SZ_V* are voxel sizes\n")
        strDef = "#define "
        for s in cnt_list:
            strNew += strDef + s + " " + str(Cnt[s]) + (s[3] == "V") * "f" + "\n"

        fpth.write_text(def_h[:i0] + strNew + def_h[i1:])
        rflg = True

    return rflg


def chck_sct_h(Cnt):
    """
    check if voxel size for scatter correction changed and adjust
    the CUDA header files accordingly.
    """
    rflg = False
    fpth = path_current / "niftypet" / "nipet" / "sct" / "src" / "sct.h"
    # pthcmpl = path.dirname(resource_filename(__name__, ''))
    sct_h = fpth.read_text()
    # get the region of keeping in synch with Python
    i0 = sct_h.find("//## start ##//")
    i1 = sct_h.find("//## end ##//")
    scth = sct_h[i0:i1]
    # list of constants which will be kept in sych from Python
    cnt_list = [
        "SS_IMX", "SS_IMY", "SS_IMZ", "SSE_IMX", "SSE_IMY", "SSE_IMZ", "NCOS", "SS_VXY", "SS_VXZ",
        "IS_VXZ", "SSE_VXY", "SSE_VXZ", "R_RING", "R_2", "IR_RING", "SRFCRS"]
    flg = False
    for i, s in enumerate(cnt_list):
        m = re.search("(?<=#define " + s + r")\s*\d*\.*\d*", scth)
        # if s[-3]=='V':
        if i < 7:
            # print(s, int(m.group(0)), Cnt[s])
            if Cnt[s] != int(m.group(0)):
                flg = True
                break
        else:
            # print(s, float(m.group(0)), Cnt[s])
            if Cnt[s] != float(m.group(0)):
                flg = True
                break

    # if flag is set then redefine the constants in the sct.h file
    if flg:
        strNew = dedent("""\
            //## start ##// constants definitions in synch with Python.   DO NOT MODIFY!\n
            // SCATTER IMAGE SIZE AND PROPERTIES
            // SS_* are used for the mu-map in scatter calculations
            // SSE_* are used for the emission image in scatter calculations
            // R_RING, R_2, IR_RING: ring radius, squared radius, inverse radius
            // NCOS is the number of samples for scatter angular sampling
            """)

        strDef = "#define "
        for i, s in enumerate(cnt_list):
            strNew += strDef + s + " " + str(Cnt[s]) + (i > 6) * "f" + "\n"

        fpth.write_text(sct_h[:i0] + strNew + sct_h[i1:])
        # sys.path.append(pthcmpl)
        rflg = True

    return rflg


def check_constants():
    """get the constants for the mMR from the resources file before
    getting the path to the local resources.py (on Linux machines it is in ~/.niftypet)"""
    resources = cs.get_resources()
    Cnt = resources.get_mmr_constants()

    sct_compile = chck_sct_h(Cnt)
    def_compile = chck_vox_h(Cnt)
    # sct_compile = False
    # def_compile = False

    if sct_compile or def_compile:
        txt = "NiftyPET constants were changed: needs CUDA compilation."
    else:
        txt = "- - . - -"

    log.info(
        dedent("""\
            ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            changed sct.h: {}
            changed def.h: {}
            ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
            {}
            ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~""").format(
            sct_compile, def_compile, txt))


cs.resources_setup(gpu=False) # install resources.py

# check and update the constants in C headers according to resources.py
check_constants()
try:
    nvcc_arches = cs.dev_setup() # update resources.py with a supported GPU device
except Exception as exc:
    nvcc_arches = []
    log.error("could not set up CUDA:\n%s", exc)

build_ver = ".".join(__version__.split('.')[:3]).split(".dev")[0]
cmake_args = [f"-DNIPET_BUILD_VERSION={build_ver}", f"-DPython3_ROOT_DIR={sys.prefix}"]
try:
    if nvcc_arches:
        cmake_args.append("-DCMAKE_CUDA_ARCHITECTURES=" + ";".join(sorted(nvcc_arches)))
except Exception as exc:
    if "sdist" not in sys.argv or any(i in sys.argv for i in ["build", "bdist", "wheel"]):
        log.warning("Import or CUDA device detection error:\n%s", exc)
for i in (Path(__file__).resolve().parent / "_skbuild").rglob("CMakeCache.txt"):
    i.write_text(re.sub("^//.*$\n^[^#].*pip-build-env.*$", "", i.read_text(), flags=re.M))
setup(use_scm_version=True, packages=find_packages(exclude=["examples", "tests"]),
      package_data={"niftypet": ["nipet/auxdata/*"]}, cmake_source_dir="niftypet",
      cmake_languages=("C", "CXX", "CUDA"), cmake_minimum_required_version="3.18",
      cmake_args=cmake_args)



"""hist.py: processing of PET list-mode data: histogramming and randoms estimation."""
import logging
import os
from collections.abc import Collection
from numbers import Integral

import nibabel as nib
import numpy as np
import scipy.ndimage as ndi

from niftypet import nimpa

from .. import mmraux

# CUDA extension module
from . import mmr_lmproc

log = logging.getLogger(__name__)

# ===============================================================================
# HISTOGRAM THE LIST-MODE DATA
# -------------------------------------------------------------------------------


def mmrhist(datain, scanner_params, t0=0, t1=0, outpath='', frms=None, use_stored=False,
            store=False, cmass_sig=5):
    '''
    Process the list-mode data and return histogram, head curves,
    and centre of mass for motion detection.
    '''
    # constants, transaxial and axial LUTs are extracted
    Cnt = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    return hist(datain, txLUT, axLUT, Cnt, frms=frms, use_stored=use_stored, store=store,
                outpath=outpath, t0=t0, t1=t1, cmass_sig=cmass_sig)


def hist(
        datain,
        txLUT,
        axLUT,
        Cnt,
        t0=0,
        t1=0,
        cmass_sig=5,
        frms=None,        # np.array([0], dtype=np.uint16),
        use_stored=False,
        store=False,
        outpath=''):
    '''
    Process list mode data with histogramming and optional bootstrapping:
    Cnt['BTP'] = 0: no bootstrapping [default];
    Cnt['BTP'] = 1: non-parametric bootstrapping;
    Cnt['BTP'] = 2: parametric bootstrapping (using Poisson distribution with mean = 1)
    '''

    if Cnt['SPN'] == 1: nsinos = Cnt['NSN1']
    elif Cnt['SPN'] == 11: nsinos = Cnt['NSN11']
    elif Cnt['SPN'] == 0: nsinos = Cnt['NSEG0']

    log.debug('histogramming with span {}.'.format(Cnt['SPN']))

    if (use_stored is True and 'sinos' in datain
            and os.path.basename(datain['sinos']) == f"sinos_s{Cnt['SPN']}_frm-{t0}-{t1}.npz"):
        hstout = dict(np.load(datain['sinos'], allow_pickle=True))
        nitag = len(hstout['phc'])
        log.debug('acquisition duration by integrating time tags is {} sec.'.format(nitag))

    elif os.path.isfile(datain['lm_bf']):
        # gather info about the LM time tags
        nele, ttags, tpos = mmr_lmproc.lminfo(datain['lm_bf'])
        nitag = int((ttags[1] - ttags[0] + 999) / 1000)
        log.debug('acquisition duration by integrating time tags is {} sec.'.format(nitag))

        # adjust frame time if outside the limit
        if t1 > nitag: t1 = nitag
        # check if the time point is allowed
        if t0 >= nitag:
            raise ValueError(
                'e> the time frame definition is not allowed! (outside acquisition time)')

        # ---------------------------------------
        # preallocate all the output arrays
        VTIME = 2
        MXNITAG = 5400 # limit to 1hr and 30mins
        if (nitag > MXNITAG):
            tn = int(MXNITAG / (1 << VTIME))
        else:
            tn = int((nitag + (1 << VTIME) - 1) / (1 << VTIME))

        pvs = np.zeros((tn, Cnt['NSEG0'], Cnt['NSBINS']), dtype=np.uint32)
        phc = np.zeros((nitag), dtype=np.uint32)
        dhc = np.zeros((nitag), dtype=np.uint32)
        mss = np.zeros((nitag), dtype=np.float32)

        bck = np.zeros((2, nitag, Cnt['NBCKT']), dtype=np.uint32)
        fan = np.zeros((Cnt['NRNG'], Cnt['NCRS']), dtype=np.uint32)

        # > prompt and delayed sinograms
        psino = np.zeros((nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.uint16)
        dsino = np.zeros((nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.uint16)

        # > single slice rebinned prompots
        ssr = np.zeros((Cnt['NSEG0'], Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.uint32)

        hstout = {
            'phc': phc, 'dhc': dhc, 'mss': mss, 'pvs': pvs, 'bck': bck, 'fan': fan, 'psn': psino,
            'dsn': dsino, 'ssr': ssr}
        # ---------------------------------------

        # do the histogramming and processing
        mmr_lmproc.hist(hstout, datain['lm_bf'], t0, t1, txLUT, axLUT, Cnt)

        if store:
            if outpath == '':
                fsino = os.path.dirname(datain['lm_bf'])
            else:
                fsino = os.path.join(outpath, 'sino')
                nimpa.create_dir(fsino)
            # complete the path with the file name
            fsino = os.path.join(fsino, f"sinos_s{Cnt['SPN']}_frm-{t0}-{t1}.npz")
            # store to the above path
            np.savez(fsino, **hstout)

    else:
        log.error('input list-mode data is not defined.')
        return

    # short (interval) projection views
    pvs_sgtl = np.right_shift(hstout['pvs'], 8).astype(np.float32)
    pvs_crnl = np.bitwise_and(hstout['pvs'], 255).astype(np.float32)

    cmass = Cnt['SO_VXZ'] * ndi.filters.gaussian_filter(hstout['mss'], cmass_sig, mode='mirror')
    log.debug(
        'centre of mass of axial radiodistribution (filtered with Gaussian of SD ={}):  COMPLETED.'
        .format(cmass_sig))

    # ========================= BUCKET SINGLES =========================
    # > number of single rates reported for the given second
    # > the last two bits are used for the number of reports
    nsr = (hstout['bck'][1, :, :] >> 30)

    # > average in a second period
    hstout['bck'][0, nsr > 0] = hstout['bck'][0, nsr > 0] / nsr[nsr > 0]

    # > time indeces when single rates given
    tmsk = np.sum(nsr, axis=1) > 0
    single_rate = np.copy(hstout['bck'][0, tmsk, :])

    # > time
    t = np.arange(nitag)
    t = t[tmsk]

    # > get the average bucket singles:
    buckets = np.int32(np.sum(single_rate, axis=0) / single_rate.shape[0])
    log.debug('dynamic and static buckets single rates:  COMPLETED.')
    # ==================================================================

    # account for the fact that when t0==t1 that means that full dataset is processed
    if t0 == t1: t1 = t0 + nitag

    return {
        't0': t0,
        't1': t1,
        'dur': t1 - t0,           # duration
        'phc': hstout['phc'],     # prompts head curve
        'dhc': hstout['dhc'],     # delayeds head curve
        'cmass': cmass,           # centre of mass of the radiodistribution in axial direction
        'pvs_sgtl': pvs_sgtl,     # sagittal projection views in short intervals
        'pvs_crnl': pvs_crnl,     # coronal projection views in short intervals
        'fansums': hstout['fan'], # fan sums of delayeds for variance reduction of randoms
        'sngl_rate': single_rate, # bucket singles over time
        'tsngl': t,               # time points of singles measurements in list-mode data
        'buckets': buckets,       # average bucket singles
        'psino': hstout['psn'].astype(np.uint16), # prompt sinogram
        'dsino': hstout['dsn'].astype(np.uint16), # delayeds sinogram
        'pssr': hstout['ssr']     # single-slice rebinned sinogram of prompts
    }  # yapf: disable


# ==============================================================================
# GET REDUCED VARIANCE RANDOMS
# ------------------------------------------------------------------------------


def randoms(hst, scanner_params, gpu_dim=False):
    '''
    Get the estimated sinogram of random events using the delayed event
    measurement.  The delayed sinogram is in the histogram dictionary
    obtained from the processing of the list-mode data.
    '''

    # constants, transaxial and axial LUTs are extracted
    Cnt = scanner_params['Cnt']
    txLUT = scanner_params['txLUT']
    axLUT = scanner_params['axLUT']

    rndsino, singles = rand(hst['fansums'], txLUT, axLUT, Cnt)

    if gpu_dim:
        rsng = mmraux.remgaps(rndsino, txLUT, Cnt)
        return rsng, singles
    else:
        return rndsino, singles


def rand(fansums, txLUT, axLUT, Cnt):
    if Cnt['SPN'] == 1: nsinos = Cnt['NSN1']
    elif Cnt['SPN'] == 11: nsinos = Cnt['NSN11']
    elif Cnt['SPN'] == 0: nsinos = Cnt['NSEG0']

    # random sino and estimated crystal map of singles put into a dictionary
    rsn = np.zeros((nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
    cmap = np.zeros((Cnt['NCRS'], Cnt['NRNG']), dtype=np.float32)
    rndout = {'rsn': rsn, 'cmap': cmap}

    mmr_lmproc.rand(rndout, fansums, txLUT, axLUT, Cnt)

    return rndout['rsn'], rndout['cmap']


# ===============================================================================
# NEW!! GET REDUCED VARIANCE RANDOMS (BASED ON PROMPTS)
# -------------------------------------------------------------------------------


def prand(fansums, pmsk, txLUT, axLUT, Cnt):
    if Cnt['SPN'] == 1: nsinos = Cnt['NSN1']
    elif Cnt['SPN'] == 11: nsinos = Cnt['NSN11']
    elif Cnt['SPN'] == 0: nsinos = Cnt['NSEG0']

    # number of frames
    nfrm = fansums.shape[0]
    log.debug('# of dynamic frames: {}.'.format(nfrm))

    # random sino and estimated crystal map of singles put into a dictionary
    rsn = np.zeros((nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
    cmap = np.zeros((Cnt['NCRS'], Cnt['NRNG']), dtype=np.float32)
    rndout = {'rsn': rsn, 'cmap': cmap}

    # save results for each frame

    rsino = np.zeros((nfrm, nsinos, Cnt['NSANGLES'], Cnt['NSBINS']), dtype=np.float32)
    crmap = np.zeros((nfrm, Cnt['NCRS'], Cnt['NRNG']), dtype=np.float32)

    for i in range(nfrm):
        rndout['rsn'][:, :, :] = 0
        rndout['cmap'][:, :] = 0
        mmr_lmproc.prand(rndout, pmsk, fansums[i, :, :], txLUT, axLUT, Cnt)
        rsino[i, :, :, :] = rndout['rsn']
        crmap[i, :, :] = rndout['cmap']

    if nfrm == 1:
        rsino = rsino[0, :, :, :]
        crmap = crmap[0, :, :]

    return rsino, crmap


def sino2nii(sino, Cnt, fpth):
    '''save sinogram in span-11 into NIfTI file'''
    # number of segments
    segn = len(Cnt['SEG'])
    cumseg = np.cumsum(Cnt['SEG'])
    cumseg = np.append([0], cumseg)

    # plane offset (relative to 127 planes of seg 0) for each segment
    OFF = np.min(abs(np.append([Cnt['MNRD']], [Cnt['MXRD']], axis=0)), axis=0)
    niisn = np.zeros((Cnt['SEG'][0], Cnt['NSANGLES'], Cnt['NSBINS'], segn), dtype=sino.dtype)

    # first segment (with direct planes)
    # tmp =
    niisn[:, :, :, 0] = sino[Cnt['SEG'][0] - 1::-1, ::-1, ::-1]

    for iseg in range(1, segn):
        niisn[OFF[iseg]:OFF[iseg] + Cnt['SEG'][iseg], :, :,
              iseg] = sino[cumseg[iseg] + Cnt['SEG'][iseg] - 1:cumseg[iseg] - 1:-1, ::-1, ::-1]

    niisn = np.transpose(niisn, (2, 1, 0, 3))

    nim = nib.Nifti1Image(niisn, np.eye(4))
    nib.save(nim, fpth)


# ================================================================================
# create michelogram map for emission data, only when the input sino in in span-1
def get_michem(sino, axLUT, Cnt):
    if Cnt['SPN'] == 1:
        slut = np.arange(Cnt['NSN1']) # for span 1, one-to-one mapping
    elif Cnt['SPN'] == 11:
        slut = axLUT['sn1_sn11']
    else:
        raise ValueError('sino is neither in span-1 or span-11')

    # acitivity michelogram
    Mem = np.zeros((Cnt['NRNG'], Cnt['NRNG']), dtype=np.float32)
    # sino to ring number & sino-1 to sino-11 index:
    sn1_rno = axLUT['sn1_rno']
    # sum all the sinograms inside
    ssm = np.sum(sino, axis=(1, 2))

    for sni in range(len(sn1_rno)):
        r0 = sn1_rno[sni, 0]
        r1 = sn1_rno[sni, 1]
        Mem[r1, r0] = ssm[slut[sni]]

    return Mem


# ================================================================================
# --------------------------------------------------------------------------------
# ================================================================================


def draw_frames(hst, tfrms, plot_diff=True):
    import matplotlib.pyplot as plt
    diff = np.int64(hst['phc']) - np.int64(hst['dhc'])
    plt.figure()
    plt.plot(hst['phc'], label='prompts')
    plt.plot(hst['dhc'], label='randoms')
    if plot_diff:
        plt.plot(diff, label='difference')

    K = [f[0] for f in tfrms if isinstance(f, list)]
    for k in K:
        yval = hst['phc'][k]
        if yval < 0.2 * np.max(hst['phc']):
            yval = 0.2 * np.max(hst['phc'])
        plt.plot([k, k], [0, yval], 'k--', lw=.75)
    plt.legend()
    plt.xlabel('time [sec]')
    plt.ylabel('counts/sec')
    plt.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))


def get_time_offset(hst):
    '''
    Detects when the signal is stronger than the randoms (noise) in the list-mode data stream.
    '''
    # detect when the signal (here prompt data) is almost as strong as randoms
    s = hst['dhc'] > 0.98 * hst['phc']
    # return index, which will constitute time in seconds, for this offset
    return len(s) - np.argmax(s[::-1]) - 1


def split_frames(hst, Tref=0, t0=0, t1=0):
    '''
    Splits the whole acquisition data into approximately statistically
    equivalent frames relative to the reference frame whose duration is
    Tref or t1-t0.  The next frames will have a similar count level.
    hst: histogram dictionary
    Tref: reference duration in seconds
    t0: start time of the reference frame
    t1: end time of the reference frame
    '''
    # get the offset
    toff = get_time_offset(hst)
    # difference between prompts and randoms
    diff = np.int64(hst['phc']) - np.int64(hst['dhc'])

    # follow up index
    i = t0 + (toff) * (t0 <= 0)
    if Tref > 0:
        j = i + Tref
    elif t1 > 0:
        j = t1 + (toff) * (t0 <= 0)
    else:
        raise ValueError('e> could not figure out the reference frame.')

    # reference count level
    cref = np.sum(diff[i:j])
    # cumulative sum of the difference
    csum = np.cumsum(diff)

    i = 0
    j = toff
    # threshold to be achieved
    thrsh = csum[j - 1] + cref
    fdur = [toff]
    frms = ['timings', [0, toff]]
    clvl = [0]
    print('frame counts t(%d,%d) = %d. diff=%d' % (i, j, clvl[-1], np.sum(diff[i:j]) - cref))
    while thrsh < csum[-1]:
        i = j
        j = np.argmax(csum > thrsh)
        fdur.append(j - i)
        frms.append([i, j])
        clvl.append(np.sum(diff[i:j]))
        print('frame counts t(%d,%d) = %d. diff=%d' % (i, j, clvl[-1], np.sum(diff[i:j]) - cref))
        thrsh += cref
    # last remianing frame
    i = j
    j = hst['dur']
    # if last frame is short, include it in the last one.
    if np.sum(diff[i:]) > .5 * cref:
        fdur.append(j - i)
        frms.append([i, j])
        clvl.append(np.sum(diff[i:]))
    else:
        fdur[-1] += j - i
        frms[-1][-1] += j - i
        clvl[-1] += np.sum(diff[i:])
        i = frms[-1][0]
    print('frame counts t(%d,%d) = %d. diff=%d' % (i, j, clvl[-1], np.sum(diff[i:j]) - cref))
    return {'timings': frms, 'fdur': fdur, 'fcnts': clvl, 'offset': toff, 'csum': csum}


def frame_position(hst, tposition, Cref=0, tr0=0, tr1=15, verbose=True):
    '''
    Inputs:
        hst: histogram data
        tposition: time position (middle point) of the frame to be defined
        Cref: reference count level to be in the frame (prompts - delays)
        tr0, tr1: time definitions of reference frame whose count level
        will be used as the reference Cref.  If Cref is not defined (i.e. = 0)
        then the tr0 and tr1 will be used.
    '''
    # claculate the difference between the prompts and delays (more realistic count level)
    diff = np.int64(hst['phc']) - np.int64(hst['dhc'])
    # cumulative sum for calculating count levels in arbitrary time windows
    cumdiff = np.cumsum(diff)

    if Cref == 0:
        Cref = cumdiff[tr1] - cumdiff[tr0 - 1]

    if Cref < 0:
        raise ValueError('The reference count level has to be non-negative')

    (log.info if verbose else log.debug)('reference count level: {}.'.format(Cref))

    stp0 = 0
    stp1 = 0
    Cw = 0
    while Cw < Cref:
        # check if it is possible to widen the sampling window both ways
        if (tposition - stp0 - 1) > 0: stp0 += 1
        if (tposition + stp1 + 1) <= len(cumdiff) - 1: stp1 += 1
        Cw = cumdiff[tposition + stp1] - cumdiff[tposition - stp0 - 1]

    tw0 = tposition - stp0
    tw1 = tposition + stp1
    Tw = tw1 - tw0
    (log.info if verbose else log.debug)(
        'time window t[{}, {}] of duration T={} and count level Cw={}'.format(tw0, tw1, Tw, Cw))

    return (tw0, tw1)


def auxilary_frames(hst, t_frms, Cref=0, tr0=0, tr1=15, verbose=True):
    '''
    Get auxiliary time frames with equal count levels for constant precision in
    the estimation of subject motion based on PET data.
    '''

    # calculate the difference between the prompts and delays (more realistic count level)
    diff = np.int64(hst['phc']) - np.int64(hst['dhc'])

    # previous frame (time tuple)
    prev_frm = (0, 0)
    # previous frame index
    prev_i = -1
    # look up table to the auxilary frames from the regular ones
    timings = []
    fi2afi = []
    for i in range(len(t_frms)):
        # time point as an average between the frame end points
        tp = int(np.mean([t_frms[i][0], t_frms[i][1]]))
        # alternative (more accurate) average through centre of mass
        t0 = t_frms[i][0]
        t1 = t_frms[i][1]
        if t1 >= hst['dur']: t1 = hst['dur'] - 1
        t = np.arange(t0, t1)
        tcm = np.sum(diff[t] * t) / np.sum(diff[t])
        # get the tuple of the equivalent count level frame
        frm = frame_position(hst, tcm, tr0=tr0, tr1=tr1, verbose=False)
        # form the LUT
        if frm != prev_frm:
            prev_frm = frm
            prev_i += 1
            timings.append(list(frm))
        fi2afi.append(prev_i)
        if verbose:
            print('t[{}, {}]; tp={}, tcm={} => frm id:{}, timings:{}'.format(
                t_frms[i][0], t_frms[i][1], tp, tcm, fi2afi[-1], timings[-1]))
    # form the list of auxilary dynamic frames of equivalent count level (as in Cref)
    mfrm = ['fluid'] + timings
    return {'timings': mfrm, 'frame_idx': fi2afi}


def dynamic_timings(flist, offset=0):
    '''
    Get start and end frame timings from a list of dynamic PET frame definitions.
    Arguments:
      flist: can be 1D list of time duration for each dynamic frame, e.g.:
            flist = [15, 15, 15, 15, 30, 30, 30, ...]
        or a 2D list of lists having 2 entries per definition:
        first for the number of repetitions and the other for the frame duration, e.g.:
            flist = ['def', [4, 15], [8, 30], ...],
        meaning 4x15s, then 8x30s, etc.
      offset: adjusts for the start time (usually when prompts are strong enough over randoms)
    Returns (dict):
      'timings': [[0, 15], [15, 30], [30, 45], [45, 60], [60, 90], [90, 120], [120, 150], ...]
      'total': total time
      'frames': array([ 15,  15,  15,  15,  30,  30,  30,  30, ...])
    '''
    if not isinstance(flist, Collection) or isinstance(flist, str):
        raise TypeError('Wrong type of frame data input')
    if all(isinstance(t, Integral) for t in flist):
        tsum = offset
        # list of frame timings
        if offset > 0:
            t_frames = ['timings', [0, offset]]
        else:
            t_frames = ['timings']
        for i in range(len(flist)):
            # frame start time
            t0 = tsum
            tsum += flist[i]
            # frame end time
            t1 = tsum
            # append the timings to the list
            t_frames.append([t0, t1])
        frms = np.uint16(flist)
    elif flist[0] == 'def' and all(isinstance(t, Collection) and len(t) == 2 for t in flist[1:]):
        flist = flist[1:]
        if offset > 0:
            flist.insert(0, [0, offset])
            farray = np.asarray(flist, dtype=np.uint16)
        else:
            farray = np.array(flist)
        # number of dynamic frames
        nfrm = np.sum(farray[:, 0])
        # list of frame duration
        frms = np.zeros(nfrm, dtype=np.uint16)
        # frame iterator
        fi = 0
        # time sum of frames
        tsum = 0
        # list of frame timings
        t_frames = ['timings']
        for i in range(farray.shape[0]):
            for _ in range(farray[i, 0]):
                # frame start time
                t0 = tsum
                tsum += farray[i, 1]
                # frame end time
                t1 = tsum
                # append the timings to the list
                t_frames.append([t0, t1])
                frms[fi] = farray[i, 1]
                fi += 1
    else:
        raise TypeError('Unrecognised time frame definitions.')
    return {'total': tsum, 'frames': frms, 'timings': t_frames}
